# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

from typing import Sequence, Union

import onnx_ir as ir

from onnxscript.rewriter import _fusion_utils, pattern

Dim = Union[int, ir.SymbolicDim]

# Fusion rule for SkipRMSNormalization


class SkipRmsNormFusion(pattern.RewriteRuleClassBase):
    def __init__(self, name: str, has_bias: bool = False, bias_pre_add: bool = False):
        """Fusion rule for SkipRMSNormalization."""
        super().__init__(name=name)
        self._has_bias = has_bias
        self._bias_pre_add = bias_pre_add

    def pattern(self, op, input, skip, gamma, bias, epsilon, stash_type):
        if self._has_bias and self._bias_pre_add:
            input = op.Add(input, bias)

        # Support different combinations of addition of input and skip
        skip_sum_pattern_1 = op.Add(skip, input)
        skip_sum_pattern_2 = op.Add(input, skip)
        skip_sum = pattern.OrValue(
            [skip_sum_pattern_1, skip_sum_pattern_2],
            name="skip_sum",
        )

        if self._has_bias and not self._bias_pre_add:
            skip_sum = op.Add(skip_sum, bias)
        # Note: ORT's SimplifiedLayerNormalization was placed in onnx domain by mistake.
        # No need to use com.microsoft domain here; but this is a custom op in ORT.
        normalized = op.SimplifiedLayerNormalization(
            skip_sum,
            gamma,
            axis=-1,
            _allow_other_attributes=True,
            _outputs=["simplified_layer_norm"],
        )
        return normalized, skip_sum

    def check(
        self,
        op,
        input,
        skip,
        gamma,
        bias,
        simplified_layer_norm,
        **_,
    ) -> pattern.MatchResult:  # type: ignore[name-defined]
        """Check if the pattern matches conditions for use of SkipSimplifiedLayerNormalization op."""
        check_result = pattern.MatchResult()
        bindings: dict[str, Dim] = {}

        def no_match(val: ir.Value, dims: Sequence[str]) -> bool:
            return not _fusion_utils.check_shape_bool(bindings, val, dims)

        if no_match(input, ["B", "S", "D"]):
            return check_result.fail(
                f"Shape mismatch: {input} does not match expected dimensions ['B', 'S', 'D']",
                input,
            )
        if no_match(skip, ["B", "S", "D"]):
            return check_result.fail(
                f"Shape mismatch: {skip} does not match expected dimensions ['B', 'S', 'D']",
                skip,
            )
        if no_match(gamma, ["D"]):
            return check_result.fail(
                f"Shape mismatch: {gamma} does not match expected dimensions ['D']",
                gamma,
            )
        if self._has_bias:
            if no_match(bias, ["D"]):
                return check_result.fail(
                    f"Shape mismatch: {bias} does not match expected dimensions ['D']",
                    bias,
                )

        stash_type = simplified_layer_norm.producer().attributes.get_int("stash_type", 1)
        if stash_type != 1:
            return check_result.fail("Stash type is not supported.")

        return check_result

    def rewrite(
        self,
        op,
        input,
        skip,
        gamma,
        bias,
        simplified_layer_norm,
        **_,
    ):
        epsilon = simplified_layer_norm.producer().attributes.get_float("epsilon", 1e-5)

        if self._has_bias:
            normalized, _mean, _inv_std_var, skip_sum = op.SkipSimplifiedLayerNormalization(
                input,
                skip,
                gamma,
                bias,
                epsilon=epsilon,
                _outputs=4,
                _domain="com.microsoft",
            )
        else:
            normalized, _mean, _inv_std_var, skip_sum = op.SkipSimplifiedLayerNormalization(
                input,
                skip,
                gamma,
                epsilon=epsilon,
                _outputs=4,
                _domain="com.microsoft",
            )
        return normalized, skip_sum


_skip_rms_add_bias_rule = SkipRmsNormFusion.rule(
    "SkipRmsNormBias", has_bias=True, bias_pre_add=False
)
_skip_rms_pre_add_bias_rule = SkipRmsNormFusion.rule(
    "SkipRmsNormPreBias", has_bias=True, bias_pre_add=True
)
_skip_rms_rule = SkipRmsNormFusion.rule("SkipRmsNorm", has_bias=False)

skip_rms_normalization_ruleset = pattern.RewriteRuleSet(
    [_skip_rms_pre_add_bias_rule, _skip_rms_add_bias_rule, _skip_rms_rule]
)
fuse_skip_rms_normalization = _fusion_utils.apply_fusion_rules(skip_rms_normalization_ruleset)


# Fusion rule for SkipLayerNormalization
class SkipLayerNormFusion(pattern.RewriteRuleClassBase):
    def __init__(self, name: str, has_bias: bool = False, bias_pre_add: bool = False):
        """Fusion rule for SkipLayerNormalization."""
        super().__init__(name=name)
        self._has_bias = has_bias
        self._bias_pre_add = bias_pre_add

    def pattern(self, op, input, skip, gamma, beta, bias):
        if self._has_bias and self._bias_pre_add:
            input = op.Add(input, bias)

        # Support different combinations of addition of input and skip
        skip_sum_pattern_1 = op.Add(skip, input)
        skip_sum_pattern_2 = op.Add(input, skip)
        skip_sum = pattern.OrValue([skip_sum_pattern_1, skip_sum_pattern_2], name="skip_sum")

        if self._has_bias and not self._bias_pre_add:
            skip_sum = op.Add(skip_sum, bias)

        normalized = op.LayerNormalization(
            skip_sum,
            gamma,
            beta,
            axis=-1,
            _allow_other_attributes=True,
            _outputs=["layer_norm"],
        )
        return normalized, skip_sum

    def check(
        self,
        op,
        input,
        skip,
        gamma,
        beta,
        bias,
        layer_norm,
        **_,
    ) -> pattern.MatchResult:  # type: ignore[name-defined]
        """Check if the pattern matches conditions for use of SimplifiedLayerNormalization op."""
        check_result = pattern.MatchResult()
        bindings: dict[str, Dim] = {}

        def no_match(val: ir.Value, dims: Sequence[str]) -> bool:
            return not _fusion_utils.check_shape_bool(bindings, val, dims)

        if no_match(input, ["B", "S", "D"]):
            return check_result.fail(
                f"Shape mismatch: {input} does not match expected dimensions ['B', 'S', 'D']",
                input,
            )
        if no_match(skip, ["B", "S", "D"]):
            return check_result.fail(
                f"Shape mismatch: {skip} does not match expected dimensions ['B', 'S', 'D']",
                skip,
            )
        if no_match(gamma, ["D"]):
            return check_result.fail(
                f"Shape mismatch: {gamma} does not match expected dimensions ['D']",
                gamma,
            )
        if no_match(beta, ["D"]):
            return check_result.fail(
                f"Shape mismatch: {beta} does not match expected dimensions ['D']",
                beta,
            )
        if self._has_bias:
            if no_match(bias, ["D"]):
                return check_result.fail(
                    f"Shape mismatch: {bias} does not match expected dimensions ['D']",
                    bias,
                )

        stash_type = layer_norm.producer().attributes.get_int("stash_type", 1)
        if stash_type != 1:
            return check_result.fail("Stash type is not supported.")
        return check_result

    def rewrite(
        self,
        op,
        input,
        skip,
        gamma,
        beta,
        bias,
        layer_norm,
        **_,
    ):
        epsilon = layer_norm.producer().attributes.get_float("epsilon", 1e-5)

        normalized, _mean, _inv_std_var, skip_sum = op.SkipLayerNormalization(
            input,
            skip,
            gamma,
            beta,
            bias,
            epsilon=epsilon,
            _outputs=4,
            _domain="com.microsoft",
        )
        return normalized, skip_sum


_skip_layer_add_bias_rule = SkipLayerNormFusion.rule(
    "SkipLayerNormBias", has_bias=True, bias_pre_add=False
)
_skip_layer_pre_add_bias_rule = SkipLayerNormFusion.rule(
    "SkipLayerNormPreBias", has_bias=True, bias_pre_add=True
)
_skip_layer_rule = SkipLayerNormFusion.rule("SkipLayerNorm", has_bias=False)

skip_layer_normalization_ruleset = pattern.RewriteRuleSet(
    [
        _skip_layer_pre_add_bias_rule,
        _skip_layer_add_bias_rule,
        _skip_layer_rule,
    ]
)

fuse_skip_layer_normalization = _fusion_utils.apply_fusion_rules(
    skip_layer_normalization_ruleset
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Unit tests for the tensor module."""

import unittest

import numpy as np

from onnxscript import tensor


class TestTensor(unittest.TestCase):
    def _check_values_and_shape(self, result, elements, shape):
        values = list(result.value.flatten())
        self.assertEqual(values, elements)
        self.assertEqual(result.shape, shape)

    def test_scalar_tensor_supports_python_range(self):
        x = tensor.Tensor(np.array(3))
        self.assertEqual(range(x), range(3))

    def test_scalar_tensor_supports_int_conversion(self):
        x = tensor.Tensor(np.array(1))
        self.assertEqual(int(x), 1)

    def test_multi_dimensional_tensor_errors_when_converted_to_int(self):
        x = tensor.Tensor(np.array([1, 2, 3]))
        with self.assertRaises(TypeError):
            int(x)

    def test_scalar_tensor_supports_bool_conversion(self):
        x = tensor.Tensor(np.array(1))
        self.assertEqual(bool(x), True)

    def test_multi_dimensional_tensor_errors_when_converted_to_float(self):
        x = tensor.Tensor(np.array([1, 2, 3]))
        with self.assertRaises(TypeError):
            float(x)

    def test_scalar_tensor_supports_float_conversion(self):
        x = tensor.Tensor(np.array(1))
        self.assertEqual(float(x), 1.0)

    def test_multi_dimensional_tensor_errors_when_converted_to_bool(self):
        x = tensor.Tensor(np.array([1, 2, 3]))
        with self.assertRaises(ValueError):
            bool(x)

    def test_it_can_be_initialized_with_none(self):
        x = tensor.Tensor(None)
        with self.assertRaises(ValueError):
            _ = x.value

    def test_rank_is_zero_for_scalar(self):
        x = tensor.Tensor(np.array(1))
        self.assertEqual(x.rank, 0)

    def test_rank_is_one_for_vector(self):
        x = tensor.Tensor(np.array([1, 2, 3]))
        self.assertEqual(x.rank, 1)

    def test_rank_is_two_for_matrix(self):
        x = tensor.Tensor(np.array([[1, 2, 3], [4, 5, 6]]))
        self.assertEqual(x.rank, 2)

    def test_getitem_index(self):
        # Create tensor:
        # [ [0, 1, 2],
        #   [3, 4, 5] ]
        data = np.array(range(6), dtype=np.int32).reshape(2, 3)
        x = tensor.Tensor(data)

        y = x[0]
        # Should return first row: [0, 1, 2]
        self._check_values_and_shape(y, [0, 1, 2], (3,))

        y = x[:, 0]
        # Should return first column: [0, 3]
        self._check_values_and_shape(y, [0, 3], (2,))

        y = x[:, -1]
        # Should return last column: [2, 5]
        self._check_values_and_shape(y, [2, 5], (2,))

        y = x[0, 0]
        # Should return first element (as a scalar)
        self._check_values_and_shape(y, [0], ())

    def test_getitem_slice(self):
        # Create tensor:
        # [ [0, 1, 2],
        #   [3, 4, 5],
        #   [6, 7, 8],
        #   [9, 10, 11] ]
        data = np.array(range(12), dtype=np.int32).reshape(4, 3)
        x = tensor.Tensor(data)

        y = x[0:2]
        # Should return first two rows
        self._check_values_and_shape(y, [0, 1, 2, 3, 4, 5], (2, 3))

        y = x[:, 0:2]
        # Should return first two columns
        self._check_values_and_shape(y, [0, 1, 3, 4, 6, 7, 9, 10], (4, 2))

        y = x[0:2, 0:2]
        # Should return first 2x2 sub-matrix
        self._check_values_and_shape(y, [0, 1, 3, 4], (2, 2))

    def test_getitem_index_and_slice(self):
        # Create tensor:
        # [ [0, 1, 2],
        #   [3, 4, 5],
        #   [6, 7, 8],
        #   [9, 10, 11] ]
        data = np.array(range(12), dtype=np.int32).reshape(4, 3)
        x = tensor.Tensor(data)

        y = x[0:2, 0]
        # Should return first two rows, first column
        self._check_values_and_shape(y, [0, 3], (2,))

        y = x[0, 0:2]
        # Should return first row, first 2 columns:
        self._check_values_and_shape(y, [0, 1], (2,))

    def test_getitem_gather(self):
        # Create tensor:
        # [ [0, 1, 2],
        #   [3, 4, 5],
        #   [6, 7, 8],
        #   [9, 10, 11] ]
        data = np.array(range(12), dtype=np.int32).reshape(4, 3)
        x = tensor.Tensor(data)

        indices_0_and_3 = tensor.Tensor(np.array([0, 3], dtype=np.int32))
        y = x[indices_0_and_3]
        # Should return row 0 and row 3
        self._check_values_and_shape(y, [0, 1, 2, 9, 10, 11], (2, 3))

        index_0 = tensor.Tensor(np.array([0], dtype=np.int32))
        y = x[index_0]
        # Should return row 0, but of shape 1x3
        self._check_values_and_shape(y, [0, 1, 2], (1, 3))

        y = x[indices_0_and_3, index_0]
        # Should return submatrix consisting of rows 0 and 3, column 0
        self._check_values_and_shape(y, [0, 9], (2, 1))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import numpy as np
import onnx_ir as ir
import onnxruntime
import packaging.version

ORT_VERSION = packaging.version.Version(onnxruntime.__version__)


def ort_run(model_name: str, model, inputs):
    providers = ["CPUExecutionProvider"]
    model_proto = ir.serde.serialize_model(model)
    options = onnxruntime.SessionOptions()
    options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_DISABLE_ALL
    session = onnxruntime.InferenceSession(
        model_proto.SerializeToString(), options, providers=providers
    )
    return session.run(None, inputs)


def assert_allclose(outputs, expected_outputs, rtol=1e-3, atol=1e-3):
    for i, (baseline_output, optimized_output) in enumerate(zip(expected_outputs, outputs)):
        try:
            np.testing.assert_equal(baseline_output.shape, optimized_output.shape)
            np.testing.assert_allclose(baseline_output, optimized_output, rtol=rtol, atol=atol)
        except AssertionError as e:
            diff_mask = ~np.isclose(baseline_output, optimized_output, rtol=rtol, atol=atol)
            diff = np.where(diff_mask, "X", " ")
            print(diff)
            print(f"Failed for output {i} with rtol={rtol} and atol={atol}\n{e}")
            raise



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import logging

from onnxscript import ir

logger = logging.getLogger(__name__)


def _convert_inputs_from_bfloat16_to_float16(value: ir.Value) -> None:
    if value.dtype != ir.DataType.BFLOAT16:
        return
    value.dtype = ir.DataType.FLOAT16
    _insert_cast_nodes_for_float16_to_bfloat16_to_inputs(value)


def _convert_outputs_from_bfloat16_to_float16(value: ir.Value) -> None:
    if value.dtype != ir.DataType.BFLOAT16:
        return
    _insert_cast_nodes_for_bfloat16_to_float16_to_outputs(value)


def _insert_cast_nodes_for_float16_to_bfloat16_to_inputs(value: ir.Value) -> None:
    user_nodes_and_indices = tuple(value.uses())

    attr = ir.AttrInt64(name="to", value=ir.DataType.BFLOAT16)
    cast = ir.Node(
        domain="",
        op_type="Cast",
        inputs=[value],
        num_outputs=1,
        attributes=[attr],
    )
    cast.outputs[0].dtype = ir.DataType.BFLOAT16
    cast.outputs[0].shape = value.shape

    for node, index in tuple(value.uses()):
        if node is cast:
            continue
        node.replace_input_with(index, cast.outputs[0])

    # NOTE: A safer way to insert the cast node is to prepend it to the first node
    # of the graph
    assert user_nodes_and_indices[0][0].graph is not None, "The node should belong to a graph"
    user_nodes_and_indices[0][0].graph[0].prepend(cast)


def _insert_cast_nodes_for_bfloat16_to_float16_to_outputs(value: ir.Value) -> None:
    node = value.producer()
    index = value.index()
    if node is None or index is None:
        logger.warning("Output value %s has no producer or index", value)
        return

    attr = ir.AttrInt64(name="to", value=ir.DataType.FLOAT16)
    cast = ir.Node(
        domain="",
        op_type="Cast",
        inputs=[node.outputs[index]],
        num_outputs=1,
        attributes=[attr],
    )
    cast.outputs[0].dtype = ir.DataType.FLOAT16
    cast.outputs[0].shape = node.outputs[index].shape
    node.append(cast)

    assert node.graph is not None, "Node graph should not be None"
    # Update graph/function outputs
    for idx, graph_or_function_output in enumerate(node.graph.outputs):
        if graph_or_function_output == node.outputs[index]:
            node.graph.outputs[idx] = cast.outputs[0]
    # Swap the output name of the node with the output name of the cast node to
    # preserve the output name in the graph
    node.outputs[index].name, cast.outputs[0].name = (
        cast.outputs[0].name,
        node.outputs[index].name,
    )


def dtype_adapter_for_bfloat16_model(model: ir.Model) -> None:
    """Adapt the model datatype if it's bfloat16.

    Because onnxruntime does not support bfloat16 as input/output datatype, we need to
    convert the bfloat16 datatype to float16. This function will convert the bfloat16
    datatype to float16 and insert Cast nodes to convert float16 to bfloat16.

    Model:
        inputs(float16) -> Cast(bfloat16) -> nodes(bfloat16) -> Cast(float16) -> outputs(float16)

    TODO: Delete this function after onnxruntime supports bfloat16.

    Args:
        model: The model to adapt.

    """
    for input in model.graph.inputs:
        _convert_inputs_from_bfloat16_to_float16(input)
    for output in model.graph.outputs:
        _convert_outputs_from_bfloat16_to_float16(output)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Stable APIs for PyTorch 2.9."""

from __future__ import annotations

__all__ = [
    "check_model",
    "convert_version",
    "get_torchlib_ops",
    "optimize",
    "save_model_with_external_data",
]

from typing import TYPE_CHECKING

from onnxscript import version_converter
from onnxscript._framework_apis.torch_2_8 import (
    check_model,
    get_torchlib_ops,
    optimize,
    save_model_with_external_data,
)

if TYPE_CHECKING:
    import onnx_ir as ir


def convert_version(model: ir.Model, target_version: int) -> ir.Model:
    """Convert the model to the specified ONNX opset version.

    Starting from PyTorch 2.9, down conversion is turned on and supported.
    """
    version_converter.convert_version(model, target_version, fallback=True)
    return model



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Common operators shared in the torchlib library."""

# mypy: disable-error-code="misc,arg-type,type-arg,valid-type,assignment,return-value"
from __future__ import annotations

from collections.abc import Sequence

import numpy.typing as npt
import onnx

import onnxscript
import onnxscript.values
from onnxscript import BOOL, INT64, ir
from onnxscript import opset18 as op
from onnxscript.function_libs.torch_lib import _constants, tensor_typing
from onnxscript.function_libs.torch_lib.tensor_typing import RealType
from onnxscript.onnx_types import COMPLEX64, COMPLEX128, DOUBLE, FLOAT, TensorType

COMPLEX64_TYPE = COMPLEX64.dtype
COMPLEX128_TYPE = COMPLEX128.dtype

DOMAIN = f"{_constants.DOMAIN}.common"

common_opset = onnxscript.values.Opset(domain=DOMAIN, version=1)


@onnxscript.script(common_opset)
def Rank(input: tensor_typing.TTensor) -> INT64:
    """Deprecated.

    NOTE: Do not remove, for backward compatibility with PyTorch < 2.10.

    Take the rank of the input tensor.
    """

    return op.Size(op.Shape(input))


@onnxscript.script(common_opset)
def IsScalar(input: tensor_typing.TTensor) -> BOOL:
    """Deprecated.

    NOTE: Do not remove, for backward compatibility with PyTorch < 2.10.

    Return whether the input has rank 0, or is a scalar.
    """

    return op.Equal(op.Size(op.Shape(input)), op.Constant(value_int=0))


def cast_to(a: RealType, dtype: int) -> RealType:
    """Cast input to dtype while handling complex types."""

    # Traced function because different if branches return different dtypes
    # which is not supported in an ONNX function
    if dtype == COMPLEX128_TYPE:
        # Cast to the real representation of the complex type
        casted = op.Cast(a, to=DOUBLE.dtype)
        # Create a complex number
        real_part = op.Unsqueeze(casted, axes=[-1])
        imag_part = op.Expand(op.Cast(0.0, to=DOUBLE.dtype), op.Shape(real_part))
        result = op.Concat(real_part, imag_part, axis=-1)
    elif dtype == COMPLEX64_TYPE:
        # Cast to the real representation of the complex type
        casted = op.Cast(a, to=FLOAT.dtype)
        # Create a complex number
        real_part = op.Unsqueeze(casted, axes=[-1])
        imag_part = op.Expand(0.0, op.Shape(real_part))
        result = op.Concat(real_part, imag_part, axis=-1)
    else:
        # Cast to real numbers
        result = op.Cast(a, to=dtype)

    return result


def constant(
    array: npt.ArrayLike | onnx.TensorProto | ir.DLPackCompatible | ir.ArrayCompatible,
    dtype: int | onnx.TensorProto.DataType | ir.DataType,
) -> TensorType:
    """Utility for creating a constant tensor.

    Args:
        array: The array to convert to a constant tensor.
        dtype: The data type of the tensor.

    Returns:
        A constant node.
    """
    return op.Constant(value=ir.tensor(array, dtype=ir.DataType(dtype)))


def merge_dims(dims: Sequence[int | INT64]) -> INT64:
    """Concatenate dimensions into a single value."""

    if not dims:
        return op.Constant(value_ints=ir.AttrInt64s("value_ints", []))

    neg_one_1d = op.Constant(value_ints=ir.AttrInt64s("value_ints", [-1]))

    result_dims = [
        op.Constant(value_ints=[d]) if isinstance(d, int) else op.Reshape(d, neg_one_1d)
        for d in dims
    ]

    # Set the output type to INT64 so op.Concat can be used
    for dim in result_dims:
        dim.dtype = ir.DataType.INT64
    return op.Concat(*result_dims, axis=0)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""A torch function library for onnxscript.

The modules are named after the torch module names for grouping:

- core
- nn
- fft
- vision
- etc.
"""



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# mypy: disable-error-code="misc,arg-type,type-arg,valid-type,assignment,return-value"
"""torch.ops.aten operators under the `nested` module.

- No inplace operators.
- All functions should not have the script() decorator. This is because
    we want to delay the compilation of the function.
"""

from __future__ import annotations

from typing import Optional

from onnxscript.onnx_types import TensorType


def aten_nested_to_padded_tensor(
    self: TensorType, padding: float, output_size: Optional[int] = None
) -> TensorType:
    """nested_to_padded_tensor(Tensor self, float padding, int[]? output_size=None) -> Tensor"""

    raise NotImplementedError()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Test for the new Pattern and PatternBase classes."""

import unittest

from onnxscript import ir
from onnxscript.rewriter import pattern
from onnxscript.rewriter._basics import MatchFailureError


class PatternTest(unittest.TestCase):
    """Test Pattern functionality."""

    def test_pattern_impl_basic_functionality(self):
        """Test that Pattern can be created and used independently."""

        def simple_pattern(op, x):
            return op.Identity(x)

        # Create a Pattern
        pattern_impl = pattern.Pattern(simple_pattern, name="SimpleIdentity")

        # Verify basic properties
        self.assertEqual(pattern_impl.name, "SimpleIdentity")
        self.assertIsNotNone(pattern_impl._target_pattern)
        self.assertIsNotNone(pattern_impl._matcher)
        self.assertIsNotNone(pattern_impl._condition_function)

    def test_pattern_impl_match_method(self):
        """Test that Pattern.match method works correctly."""

        def identity_pattern(op, x):
            return op.Identity(x)

        pattern_impl = pattern.Pattern(identity_pattern, name="IdentityPattern")

        # Create a model with an Identity node
        model = ir.from_onnx_text(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x) => (float[N] z)
            {
                z = Identity(x)
            }
            """
        )

        # Find the Identity node
        identity_node = None
        for node in model.graph:
            if node.op_type == "Identity":
                identity_node = node
                break

        self.assertIsNotNone(identity_node)

        # Test pattern matching
        match_result = pattern_impl.match(model, model.graph, identity_node)

        # The match might succeed or fail depending on how the pattern matching works
        # The important thing is that the method runs without error
        self.assertIsInstance(match_result, (pattern.MatchResult, type(None)))

    def test_pattern_impl_with_condition_function(self):
        """Test Pattern with a custom condition function."""

        def identity_pattern(op, x):
            return op.Identity(x)

        def always_fail_condition(context, x):
            return False

        pattern_impl = pattern.Pattern(
            identity_pattern, condition_function=always_fail_condition, name="FailingIdentity"
        )

        # Create a model with an Identity node
        model = ir.from_onnx_text(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x) => (float[N] z)
            {
                z = Identity(x)
            }
            """
        )

        # Find the Identity node
        identity_node = None
        for node in model.graph:
            if node.op_type == "Identity":
                identity_node = node
                break

        self.assertIsNotNone(identity_node)

        # Test pattern matching - should fail due to condition function
        match_result = pattern_impl.match(model, model.graph, identity_node)

        # Should return None due to failing condition
        self.assertIsNone(match_result)

    def test_pattern_impl_no_match_returns_match_object(self):
        """Test that Pattern.match returns match object (not always None) when available."""

        def identity_pattern(op, x):
            return op.Identity(x)

        pattern_impl = pattern.Pattern(identity_pattern, name="IdentityPattern")

        # Create a model with an Add node (should not match Identity pattern)
        model = ir.from_onnx_text(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x, float[N] y) => (float[N] z)
            {
                z = Add(x, y)
            }
            """
        )

        # Find the Add node
        add_node = None
        for node in model.graph:
            if node.op_type == "Add":
                add_node = node
                break

        self.assertIsNotNone(add_node)

        # Test pattern matching - should fail because Add != Identity
        match_result = pattern_impl.match(model, model.graph, add_node)

        # The result should be falsy (either None or a failed MatchResult)
        self.assertFalse(bool(match_result))


class PatternBaseTest(unittest.TestCase):
    """Test PatternBase functionality."""

    def test_pattern_base_creation(self):
        """Test that PatternBase can be subclassed and used."""

        class TestPattern(pattern.PatternBase):
            def pattern(self, op, x):
                return op.Identity(x)

        test_pattern = TestPattern(name="TestPattern")
        self.assertEqual(test_pattern.name, "TestPattern")

    def test_pattern_base_compiled_pattern_access(self):
        """Test that PatternBase has an internal Pattern that is created on demand."""

        class TestPattern(pattern.PatternBase):
            def pattern(self, op, x):
                return op.Identity(x)

            def check(self, context, x):
                return pattern.MatchResult()  # Always succeeds

        test_pattern = TestPattern(name="TestPattern")

        # Initially, the Pattern should not be created (lazy initialization)
        self.assertIsNone(test_pattern._compiled_pattern)

        # Create a simple model to trigger pattern creation
        model = ir.from_onnx_text(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x) => (float[N] z)
            {
                z = Identity(x)
            }
            """
        )
        graph = model.graph
        node = next(iter(graph))

        # Calling match() should trigger the creation of _compiled_pattern
        test_pattern.match(model, graph, node)

        # Now the Pattern should be created
        self.assertIsInstance(test_pattern._compiled_pattern, pattern.Pattern)
        self.assertEqual(test_pattern._compiled_pattern.name, "TestPattern")

    def test_pattern_base_default_name(self):
        """Test that PatternBase uses class name as default."""

        class MyCustomPattern(pattern.PatternBase):
            def pattern(self, op, x):
                return op.Identity(x)

        test_pattern = MyCustomPattern()
        self.assertEqual(test_pattern.name, "MyCustomPattern")


class RewriteRuleInheritanceTest(unittest.TestCase):
    """Test that RewriteRule still works after inheriting from Pattern."""

    def test_rewrite_rule_still_works(self):
        """Test that existing RewriteRule functionality is preserved."""

        def reciprocal_mul_pattern(op, x, y):
            return (1 / x) * y

        def div_replacement(op, x, y):
            return op.Div(y, x)

        rule = pattern.RewriteRule(reciprocal_mul_pattern, div_replacement)

        # Create a model that should match
        model = ir.from_onnx_text(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x, float[N] y) => (float[N] z)
            {
                c1 = Constant<value_float = 1.0>()
                t1 = Div(c1, x)
                z1 = Mul(t1, y)
                z = Identity(z1)
            }
            """
        )

        # Apply the rule
        count = rule.apply_to_model(model)

        # The rule should either apply or not, but the method should work
        self.assertIsInstance(count, int)
        self.assertGreaterEqual(count, 0)

    def test_rewrite_rule_class_base_still_works(self):
        """Test that RewriteRuleClassBase still works after inheriting from PatternBase."""

        class SimpleIdentityRule(pattern.RewriteRuleClassBase):
            def pattern(self, op, x):
                return op.Identity(x)

            def check(self, context, x):
                return pattern.MatchResult()  # Always succeeds

            def rewrite(self, op, x):
                return op.Identity(x)  # No-op replacement

        # Create a rule instance
        rule = SimpleIdentityRule.rule()

        self.assertIsInstance(rule, pattern.RewriteRule)
        self.assertEqual(rule.name, "SimpleIdentityRule")


class RewriteFailureConventionsTest(unittest.TestCase):
    """Test that rewrite functions support the same failure conventions as check functions.

    The check side supports three failure conventions:
    - Return a falsy MatchResult (via .fail())
    - Raise MatchFailureError
    - Return None or False

    The rewrite side should support all three as well, for uniformity.
    """

    _IDENTITY_MODEL_TEXT = """
        <ir_version: 7, opset_import: [ "" : 17]>
        agraph (float[N] x) => (float[N] z)
        {
            z = Identity(x)
        }
    """

    def _apply_rewrite_rule(self, rewrite_fn):
        """Helper that builds a RewriteRule with the given rewrite function and applies it."""

        def identity_pattern(op, x):
            return op.Identity(x)

        rule = pattern.RewriteRule(identity_pattern, rewrite_fn, name="TestRule")
        model = ir.from_onnx_text(self._IDENTITY_MODEL_TEXT)
        count = rule.apply_to_model(model)
        return count

    def test_rewrite_returning_none_is_treated_as_failure(self):
        """Returning None from rewrite indicates failure (no replacement made)."""

        def rewrite_returns_none(op, x):
            return None

        count = self._apply_rewrite_rule(rewrite_returns_none)
        self.assertEqual(count, 0)

    def test_rewrite_returning_false_is_treated_as_failure(self):
        """Returning False from rewrite indicates failure (deprecated but supported)."""

        def rewrite_returns_false(op, x):
            return False

        count = self._apply_rewrite_rule(rewrite_returns_false)
        self.assertEqual(count, 0)

    def test_rewrite_raising_match_failure_error_is_treated_as_failure(self):
        """Raising MatchFailureError from rewrite indicates failure."""

        def rewrite_raises_error(op, x):
            raise MatchFailureError("Cannot rewrite this node")

        count = self._apply_rewrite_rule(rewrite_raises_error)
        self.assertEqual(count, 0)

    def test_rewrite_returning_falsy_match_result_is_treated_as_failure(self):
        """Returning a falsy MatchResult from rewrite indicates failure."""

        def rewrite_returns_failed_match_result(op, x):
            result = pattern.MatchResult()
            return result.fail("Rewrite not applicable")

        count = self._apply_rewrite_rule(rewrite_returns_failed_match_result)
        self.assertEqual(count, 0)

    def test_rewrite_returning_ir_value_succeeds(self):
        """Returning an ir.Value from rewrite is success (the normal case)."""

        # Use a non-self-referential pattern to avoid infinite rewrite loops
        def add_zero_pattern(op, x):
            return op.Add(x, 0.0)

        def identity_replacement(op, x):
            return op.Identity(x)

        rule = pattern.RewriteRule(add_zero_pattern, identity_replacement, name="TestRule")
        model = ir.from_onnx_text(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[N] x) => (float[N] z)
            {
                c0 = Constant<value_float = 0.0>()
                z = Add(x, c0)
            }
            """
        )
        count = rule.apply_to_model(model)
        self.assertGreaterEqual(count, 1)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
__all__ = [
    "add_0_rule",
    "affine_conv_fusion_rule",
    "cast_cast_rule",
    "cast_constant_of_shape_rule",
    "cast_constant_of_shape_without_value_rule",
    "collapse_slice_rule",
    "collapse_slice2_rule",
    "conv_affine_fusion_rule",
    "div_by_1_rule",
    "dropout_inference_rule",
    "dropout_zero_rule",
    "expand_before_binary_op_rules",
    "flatten_to_reshape_rule",
    "fuse_batchnorm_into_conv_rule",
    "fuse_batchnorm_into_conv_transpose_rule",
    "fuse_batchnorm_into_gemm_rule",
    "fuse_hardswish_rules",
    "fuse_pad_into_conv_integer_rule",
    "fuse_pad_into_conv_rule",
    "min_min_rule",
    "max_max_rule",
    "min_max_rule",
    "max_min_rule",
    "gemm_to_matmul_add_rule",
    "matmul_add_to_gemm_rule",
    "mul_by_1_rule",
    "no_op_cast_rule",
    "no_op_dynamic_scatter_nd_rule",
    "no_op_expand_rule",
    "no_op_static_scatter_nd_rule",
    "no_op_transpose_rule",
    "normalize_pad_format_conv_integer_rule",
    "normalize_pad_format_conv_rule",
    "one_reshape_matmul_reshape_rule",
    "remove_optional_bias_from_conv_rule",
    "remove_optional_bias_from_conv_transpose_rule",
    "remove_optional_bias_from_gemm_rule",
    "remove_optional_bias_from_qlinear_conv_rule",
    "reshape_reshape_rule",
    "slice_split_rule",
    "squeeze_reshape_1d_rule",
    "sub_0_rule",
    "successive_clip_relu_rule",
    "successive_clip_rule",
    "successive_relu_clip_rule",
    "successive_relu_rule",
    "transpose_a_matmul_add_to_gemm_rule",
    "transpose_ab_matmul_add_to_gemm_rule",
    "transpose_b_matmul_add_to_gemm_rule",
    "transpose_transpose_rule",
    "two_reshapes_matmul_reshape_rule",
    "unsqueeze_unsqueeze_rule",
]

from onnxscript.rewriter.rules.common._basic_rules import (
    cast_cast_rule,
    flatten_to_reshape_rule,
    no_op_cast_rule,
    no_op_expand_rule,
    no_op_transpose_rule,
    reshape_reshape_rule,
    slice_split_rule,
    squeeze_reshape_1d_rule,
    transpose_transpose_rule,
    unsqueeze_unsqueeze_rule,
)
from onnxscript.rewriter.rules.common._broadcast_to_matmul import (
    one_reshape_matmul_reshape_rule,
    two_reshapes_matmul_reshape_rule,
)
from onnxscript.rewriter.rules.common._cast_constant_of_shape import (
    cast_constant_of_shape_rule,
    cast_constant_of_shape_without_value_rule,
)
from onnxscript.rewriter.rules.common._collapse_slices import (
    collapse_slice2_rule,
    collapse_slice_rule,
)
from onnxscript.rewriter.rules.common._fuse_batchnorm import (
    fuse_batchnorm_into_conv_rule,
    fuse_batchnorm_into_conv_transpose_rule,
    fuse_batchnorm_into_gemm_rule,
)
from onnxscript.rewriter.rules.common._fuse_conv_affine import (
    affine_conv_fusion_rule,
    conv_affine_fusion_rule,
)
from onnxscript.rewriter.rules.common._fuse_hardswish import fuse_hardswish_rules
from onnxscript.rewriter.rules.common._fuse_pad_into_conv import (
    fuse_pad_into_conv_integer_rule,
    fuse_pad_into_conv_rule,
    normalize_pad_format_conv_integer_rule,
    normalize_pad_format_conv_rule,
)
from onnxscript.rewriter.rules.common._fuse_relus_clips import (
    successive_clip_relu_rule,
    successive_clip_rule,
    successive_relu_clip_rule,
    successive_relu_rule,
)
from onnxscript.rewriter.rules.common._gemm_to_matmul_add import gemm_to_matmul_add_rule
from onnxscript.rewriter.rules.common._matmul_add_to_gemm import (
    matmul_add_to_gemm_rule,
    transpose_a_matmul_add_to_gemm_rule,
    transpose_ab_matmul_add_to_gemm_rule,
    transpose_b_matmul_add_to_gemm_rule,
)
from onnxscript.rewriter.rules.common._min_max_to_clip import (
    max_max_rule,
    max_min_rule,
    min_max_rule,
    min_min_rule,
)
from onnxscript.rewriter.rules.common._no_op import (
    add_0_rule,
    div_by_1_rule,
    dropout_inference_rule,
    dropout_zero_rule,
    mul_by_1_rule,
    sub_0_rule,
)
from onnxscript.rewriter.rules.common._redundant_scatter_nd import (
    no_op_dynamic_scatter_nd_rule,
    no_op_static_scatter_nd_rule,
)
from onnxscript.rewriter.rules.common._remove_expand_before_binary_op import (
    expand_before_binary_op_rules,
)
from onnxscript.rewriter.rules.common._remove_optional_bias import (
    remove_optional_bias_from_conv_rule,
    remove_optional_bias_from_conv_transpose_rule,
    remove_optional_bias_from_gemm_rule,
    remove_optional_bias_from_qlinear_conv_rule,
)



# --------------------------------------------------------------------------
# ⚠️ WARNING - AUTO-GENERATED CODE - DO NOT EDIT ⚠️
# ⚙️ Generated by 'python -m opgen'
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# pylint: disable=W0221,W0222,R0901,W0237
# mypy: disable-error-code=override
# ruff: noqa: D402
# --------------------------------------------------------------------------

from __future__ import annotations

from typing import Optional, Sequence, Tuple, TypeVar

from onnx.defs import get_schema
from typing_extensions import TypeAlias

from onnxscript.onnx_opset._impl.opset9 import Opset9
from onnxscript.onnx_types import (
    BOOL,
    COMPLEX64,
    COMPLEX128,
    DOUBLE,
    FLOAT,
    FLOAT16,
    INT8,
    INT16,
    INT32,
    INT64,
    STRING,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
)
from onnxscript.values import Op, Opset


class Opset10(Opset9):
    def __new__(cls):
        return Opset.__new__(cls, "", 10)

    T_AveragePool = TypeVar("T_AveragePool", DOUBLE, FLOAT, FLOAT16)

    def AveragePool(
        self,
        X: T_AveragePool,
        *,
        auto_pad: str = "NOTSET",
        ceil_mode: int = 0,
        count_include_pad: int = 0,
        kernel_shape: Sequence[int],
        pads: Optional[Sequence[int]] = None,
        strides: Optional[Sequence[int]] = None,
    ) -> T_AveragePool:
        r"""[🌐 AveragePool(10)](https://onnx.ai/onnx/operators/onnx__AveragePool.html#averagepool-10 "Online Documentation")


         AveragePool consumes an input tensor X and applies average pooling across
         the tensor according to kernel sizes, stride sizes, and pad lengths.
         average pooling consisting of computing the average on all values of a
         subset of the input tensor according to the kernel size and downsampling the
         data into the output tensor Y for further processing. The output spatial shape will be following:
         ```
         output_spatial_shape[i] = floor((input_spatial_shape[i] + pad_shape[i] - kernel_spatial_shape[i]) / strides_spatial_shape[i] + 1)
         ```
         or
         ```
         output_spatial_shape[i] = ceil((input_spatial_shape[i] + pad_shape[i] - kernel_spatial_shape[i]) / strides_spatial_shape[i] + 1)
         ```
         if ceil_mode is enabled

         ```
         * pad_shape[i] is sum of pads along axis i
         ```

         `auto_pad` is a DEPRECATED attribute. If you are using them currently, the output spatial shape will be following:
         ```
         VALID: output_spatial_shape[i] = ceil((input_spatial_shape[i] - kernel_spatial_shape[i] + 1) / strides_spatial_shape[i])
         SAME_UPPER or SAME_LOWER: output_spatial_shape[i] = ceil(input_spatial_shape[i] / strides_spatial_shape[i])
         ```
         And pad shape will be following if `SAME_UPPER` or `SAME_LOWER`:
         ```
         pad_shape[i] = (output_spatial_shape[i] - 1) * strides_spatial_shape[i] + kernel_spatial_shape[i] - input_spatial_shape[i]
         ```
         The output of each pooling window is divided by the number of elements (exclude pad when attribute count_include_pad is zero).


        Args:
            X: Input data tensor from the previous operator; dimensions for image case
                are (N x C x H x W), where N is the batch size, C is the number of
                channels, and H and W are the height and the width of the data. For non
                image case, the dimensions are in the form of (N x C x D1 x D2 ... Dn),
                where N is the batch size. Optionally, if dimension denotation is in
                effect, the operation expects the input data tensor to arrive with the
                dimension denotation of [DATA_BATCH, DATA_CHANNEL, DATA_FEATURE,
                DATA_FEATURE ...].

            auto_pad: auto_pad must be either NOTSET, SAME_UPPER, SAME_LOWER or VALID.
                Where default value is NOTSET, which means explicit padding is used.
                SAME_UPPER or SAME_LOWER mean pad the input so that the output spatial
                size match the input.In case of odd number add the extra padding at the
                end for SAME_UPPER and at the beginning for SAME_LOWER. VALID mean no
                padding.

            ceil_mode: Whether to use ceil or floor (default) to compute the output
                shape.

            count_include_pad: Whether include pad pixels when calculating values for
                the edges. Default is 0, doesn't count include pad.

            kernel_shape: The size of the kernel along each axis.

            pads: Padding for the beginning and ending along each spatial axis, it can
                take any value greater than or equal to 0. The value represent the
                number of pixels added to the beginning and end part of the
                corresponding axis. `pads` format should be as follow [x1_begin,
                x2_begin...x1_end, x2_end,...], where xi_begin the number of pixels
                added at the beginning of axis `i` and xi_end, the number of pixels
                added at the end of axis `i`. This attribute cannot be used
                simultaneously with auto_pad attribute. If not present, the padding
                defaults to 0 along start and end of each spatial axis.

            strides: Stride along each spatial axis.
        """

        schema = get_schema("AveragePool", 10, "")
        op = Op(self, "AveragePool", schema)
        return op(
            *self._prepare_inputs(schema, X),
            auto_pad=auto_pad,
            ceil_mode=ceil_mode,
            count_include_pad=count_include_pad,
            kernel_shape=kernel_shape,
            pads=pads,
            strides=strides,
        )

    T1_ConvInteger = TypeVar("T1_ConvInteger", INT8, UINT8)

    T2_ConvInteger = TypeVar("T2_ConvInteger", INT8, UINT8)

    T3_ConvInteger: TypeAlias = INT32

    def ConvInteger(
        self,
        x: T1_ConvInteger,
        w: T2_ConvInteger,
        x_zero_point: Optional[T1_ConvInteger] = None,
        w_zero_point: Optional[T2_ConvInteger] = None,
        *,
        auto_pad: str = "NOTSET",
        dilations: Optional[Sequence[int]] = None,
        group: int = 1,
        kernel_shape: Optional[Sequence[int]] = None,
        pads: Optional[Sequence[int]] = None,
        strides: Optional[Sequence[int]] = None,
    ) -> T3_ConvInteger:
        r"""[🌐 ConvInteger(10)](https://onnx.ai/onnx/operators/onnx__ConvInteger.html#convinteger-10 "Online Documentation")


        The integer convolution operator consumes an input tensor, its zero-point, a filter, and its zero-point,
        and computes the output. The production MUST never overflow. The accumulation may overflow if and only if in 32 bits.


        Args:
            x: Input data tensor from previous layer; has size (N x C x H x W), where N
                is the batch size, C is the number of channels, and H and W are the
                height and width. Note that this is for the 2D image. Otherwise the size
                is (N x C x D1 x D2 ... x Dn). Optionally, if dimension denotation is in
                effect, the operation expects input data tensor to arrive with the
                dimension denotation of [DATA_BATCH, DATA_CHANNEL, DATA_FEATURE,
                DATA_FEATURE ...].

            w: The weight tensor that will be used in the convolutions; has size (M x
                C/group x kH x kW), where C is the number of channels, and kH and kW are
                the height and width of the kernel, and M is the number of feature maps.
                For more than 2 dimensions, the kernel shape will be (M x C/group x k1 x
                k2 x ... x kn), where (k1 x k2 x ... kn) is the dimension of the kernel.
                Optionally, if dimension denotation is in effect, the operation expects
                the weight tensor to arrive with the dimension denotation of
                [FILTER_OUT_CHANNEL, FILTER_IN_CHANNEL, FILTER_SPATIAL, FILTER_SPATIAL
                ...]. X.shape[1] == (W.shape[1] * group) == C (assuming zero based
                indices for the shape array). Or in other words FILTER_IN_CHANNEL should
                be equal to DATA_CHANNEL.

            x_zero_point: (optional) Zero point tensor for input 'x'. It's optional and
                default value is 0. It's a scalar, which means a per-tensor/layer
                quantization.

            w_zero_point: (optional) Zero point tensor for input 'w'. It's optional and
                default value is 0.  It could be a scalar or a 1-D tensor, which means a
                per-tensor/layer or per output channel quantization. If it's a 1-D
                tensor, its number of elements should be equal to the number of output
                channels (M)

            auto_pad: auto_pad must be either NOTSET, SAME_UPPER, SAME_LOWER or VALID.
                Where default value is NOTSET, which means explicit padding is used.
                SAME_UPPER or SAME_LOWER mean pad the input so that `output_shape[i] =
                ceil(input_shape[i] / strides[i])` for each axis `i`. The padding is
                split between the two sides equally or almost equally (depending on
                whether it is even or odd). In case the padding is an odd number, the
                extra padding is added at the end for SAME_UPPER and at the beginning
                for SAME_LOWER.

            dilations: dilation value along each spatial axis of the filter. If not
                present, the dilation defaults to 1 along each axis.

            group: number of groups input channels and output channels are divided into.
                default is 1.

            kernel_shape: The shape of the convolution kernel. If not present, should be
                inferred from input 'w'.

            pads: Padding for the beginning and ending along each spatial axis, it can
                take any value greater than or equal to 0.The value represent the number
                of pixels added to the beginning and end part of the corresponding
                axis.`pads` format should be as follow [x1_begin, x2_begin...x1_end,
                x2_end,...], where xi_begin the number ofpixels added at the beginning
                of axis `i` and xi_end, the number of pixels added at the end of axis
                `i`.This attribute cannot be used simultaneously with auto_pad
                attribute. If not present, the padding defaultsto 0 along start and end
                of each spatial axis.

            strides: Stride along each spatial axis. If not present, the stride defaults
                to 1 along each axis.
        """

        schema = get_schema("ConvInteger", 10, "")
        op = Op(self, "ConvInteger", schema)
        return op(
            *self._prepare_inputs(schema, x, w, x_zero_point, w_zero_point),
            auto_pad=auto_pad,
            dilations=dilations,
            group=group,
            kernel_shape=kernel_shape,
            pads=pads,
            strides=strides,
        )

    T_DequantizeLinear = TypeVar("T_DequantizeLinear", INT32, INT8, UINT8)

    def DequantizeLinear(
        self,
        x: T_DequantizeLinear,
        x_scale: FLOAT,
        x_zero_point: Optional[T_DequantizeLinear] = None,
    ) -> FLOAT:
        r"""[🌐 DequantizeLinear(10)](https://onnx.ai/onnx/operators/onnx__DequantizeLinear.html#dequantizelinear-10 "Online Documentation")


        The linear dequantization operator. It consumes a quantized tensor, a scale, a zero point to compute the full precision tensor.
        The dequantization formula is y = (x - x_zero_point) * x_scale. 'x_scale' and 'x_zero_point' are both scalars.
        'x_zero_point' and 'x' must have same type. 'x' and 'y' must have same shape. In the case of dequantizing int32,
        there's no zero point (zero point is supposed to be 0).


        Args:
            x: N-D quantized input tensor to be de-quantized.

            x_scale: Scale for input 'x'. It's a scalar, which means a per-tensor/layer
                quantization.

            x_zero_point: (optional) Zero point for input 'x'. It's a scalar, which
                means a per-tensor/layer quantization. It's optional. 0 is the default
                value when it's not specified.
        """

        schema = get_schema("DequantizeLinear", 10, "")
        op = Op(self, "DequantizeLinear", schema)
        return op(*self._prepare_inputs(schema, x, x_scale, x_zero_point))

    T_Dropout = TypeVar("T_Dropout", DOUBLE, FLOAT, FLOAT16)

    T1_Dropout: TypeAlias = BOOL

    def Dropout(self, data: T_Dropout, *, ratio: float = 0.5) -> Tuple[T_Dropout, T1_Dropout]:
        r"""[🌐 Dropout(10)](https://onnx.ai/onnx/operators/onnx__Dropout.html#dropout-10 "Online Documentation")


        Dropout takes one input floating tensor and produces two tensor outputs,
        output (floating tensor) and mask (`Tensor<bool>`). Depending on whether it is
        in test mode or not, the output Y will either be a random dropout, or a simple
        copy of the input. Note that our implementation of Dropout does scaling in
        the training phase, so during testing nothing needs to be done.
        This operator has **optional** inputs/outputs. See `ONNX <https://github.com/onnx/onnx/blob/master/docs/IR.md>`_ for more details about the representation of optional arguments. An empty string may be used in the place of an actual argument's name to indicate a missing argument. Trailing optional arguments (those not followed by an argument that is present) may also be simply omitted.


        Args:
            data: The input data as Tensor.

            ratio: The ratio of random dropout
        """

        schema = get_schema("Dropout", 10, "")
        op = Op(self, "Dropout", schema)
        return op(*self._prepare_inputs(schema, data), ratio=ratio)

    T1_IsInf = TypeVar("T1_IsInf", DOUBLE, FLOAT)

    T2_IsInf: TypeAlias = BOOL

    def IsInf(
        self, X: T1_IsInf, *, detect_negative: int = 1, detect_positive: int = 1
    ) -> T2_IsInf:
        r"""[🌐 IsInf(10)](https://onnx.ai/onnx/operators/onnx__IsInf.html#isinf-10 "Online Documentation")

        Map infinity to true and other values to false.

        Args:
            X: (non-differentiable) input

            detect_negative: (Optional) Whether map negative infinity to true. Default
                to 1 so that negative infinity induces true. Set this attribute to 0 if
                negative infinity should be mapped to false.

            detect_positive: (Optional) Whether map positive infinity to true. Default
                to 1 so that positive infinity induces true. Set this attribute to 0 if
                positive infinity should be mapped to false.
        """

        schema = get_schema("IsInf", 10, "")
        op = Op(self, "IsInf", schema)
        return op(
            *self._prepare_inputs(schema, X),
            detect_negative=detect_negative,
            detect_positive=detect_positive,
        )

    T1_MatMulInteger = TypeVar("T1_MatMulInteger", INT8, UINT8)

    T2_MatMulInteger = TypeVar("T2_MatMulInteger", INT8, UINT8)

    T3_MatMulInteger: TypeAlias = INT32

    def MatMulInteger(
        self,
        A: T1_MatMulInteger,
        B: T2_MatMulInteger,
        a_zero_point: Optional[T1_MatMulInteger] = None,
        b_zero_point: Optional[T2_MatMulInteger] = None,
    ) -> T3_MatMulInteger:
        r"""[🌐 MatMulInteger(10)](https://onnx.ai/onnx/operators/onnx__MatMulInteger.html#matmulinteger-10 "Online Documentation")


        Matrix product that behaves like [numpy.matmul](https://numpy.org/doc/stable/reference/generated/numpy.matmul.html).
        The production MUST never overflow. The accumulation may overflow if and only if in 32 bits.


        Args:
            A: (non-differentiable) N-dimensional matrix A

            B: (non-differentiable) N-dimensional matrix B

            a_zero_point: (optional, non-differentiable) Zero point tensor for input
                'A'. It's optional and default value is 0. It could be a scalar or N-D
                tensor. Scalar refers to per tensor quantization whereas N-D refers to
                per row quantization. If the input is 2D of shape [M, K] then zero point
                tensor may be an M element vector [zp_1, zp_2, ..., zp_M]. If the input
                is N-D tensor with shape [D1, D2, M, K] then zero point tensor may have
                shape [D1, D2, M, 1].

            b_zero_point: (optional, non-differentiable) Zero point tensor for input
                'B'. It's optional and default value is 0. It could be a scalar or a N-D
                tensor, Scalar refers to per tensor quantization whereas N-D refers to
                per col quantization. If the input is 2D of shape [K, N] then zero point
                tensor may be an N element vector [zp_1, zp_2, ..., zp_N]. If the input
                is N-D tensor with shape [D1, D2, K, N] then zero point tensor may have
                shape [D1, D2, 1, N].
        """

        schema = get_schema("MatMulInteger", 10, "")
        op = Op(self, "MatMulInteger", schema)
        return op(*self._prepare_inputs(schema, A, B, a_zero_point, b_zero_point))

    T_MaxPool = TypeVar("T_MaxPool", DOUBLE, FLOAT, FLOAT16)

    I_MaxPool: TypeAlias = INT64

    def MaxPool(
        self,
        X: T_MaxPool,
        *,
        auto_pad: str = "NOTSET",
        ceil_mode: int = 0,
        dilations: Optional[Sequence[int]] = None,
        kernel_shape: Sequence[int],
        pads: Optional[Sequence[int]] = None,
        storage_order: int = 0,
        strides: Optional[Sequence[int]] = None,
    ) -> Tuple[T_MaxPool, I_MaxPool]:
        r"""[🌐 MaxPool(10)](https://onnx.ai/onnx/operators/onnx__MaxPool.html#maxpool-10 "Online Documentation")


         MaxPool consumes an input tensor X and applies max pooling across
         the tensor according to kernel sizes, stride sizes, and pad lengths.
         max pooling consisting of computing the max on all values of a
         subset of the input tensor according to the kernel size and downsampling the
         data into the output tensor Y for further processing. The output spatial shape will be following:
         ```
         output_spatial_shape[i] = floor((input_spatial_shape[i] + pad_shape[i] - ((kernel_spatial_shape[i] - 1) * dilations[i] + 1)) / strides_spatial_shape[i] + 1)
         ```
         or
         ```
         output_spatial_shape[i] = ceil((input_spatial_shape[i] + pad_shape[i] - ((kernel_spatial_shape[i] - 1) * dilations[i] + 1)) / strides_spatial_shape[i] + 1)
         ```
         if ceil_mode is enabled

         ```
         * pad_shape[i] is sum of pads along axis i
         ```

         `auto_pad` is a DEPRECATED attribute. If you are using them currently, the output spatial shape will be following:
         ```
         VALID: output_spatial_shape[i] = ceil((input_spatial_shape[i] - ((kernel_spatial_shape[i] - 1) * dilations[i] + 1) + 1) / strides_spatial_shape[i])
         SAME_UPPER or SAME_LOWER: output_spatial_shape[i] = ceil(input_spatial_shape[i] / strides_spatial_shape[i])
         ```
         And pad shape will be following if `SAME_UPPER` or `SAME_LOWER`:
         ```
         pad_shape[i] = (output_spatial_shape[i] - 1) * strides_spatial_shape[i] + ((kernel_spatial_shape[i] - 1) * dilations[i] + 1) - input_spatial_shape[i]
         ```
         The output of each pooling window is maximum number of elements exclude pad.


        Args:
            X: Input data tensor from the previous operator; dimensions for image case
                are (N x C x H x W), where N is the batch size, C is the number of
                channels, and H and W are the height and the width of the data. For non
                image case, the dimensions are in the form of (N x C x D1 x D2 ... Dn),
                where N is the batch size. Optionally, if dimension denotation is in
                effect, the operation expects the input data tensor to arrive with the
                dimension denotation of [DATA_BATCH, DATA_CHANNEL, DATA_FEATURE,
                DATA_FEATURE ...].

            auto_pad: auto_pad must be either NOTSET, SAME_UPPER, SAME_LOWER or VALID.
                Where default value is NOTSET, which means explicit padding is used.
                SAME_UPPER or SAME_LOWER mean pad the input so that the output spatial
                size match the input.In case of odd number add the extra padding at the
                end for SAME_UPPER and at the beginning for SAME_LOWER. VALID mean no
                padding.

            ceil_mode: Whether to use ceil or floor (default) to compute the output
                shape.

            dilations: Dilation value along each spatial axis of filter.

            kernel_shape: The size of the kernel along each axis.

            pads: Padding for the beginning and ending along each spatial axis, it can
                take any value greater than or equal to 0. The value represent the
                number of pixels added to the beginning and end part of the
                corresponding axis. `pads` format should be as follow [x1_begin,
                x2_begin...x1_end, x2_end,...], where xi_begin the number of pixels
                added at the beginning of axis `i` and xi_end, the number of pixels
                added at the end of axis `i`. This attribute cannot be used
                simultaneously with auto_pad attribute. If not present, the padding
                defaults to 0 along start and end of each spatial axis.

            storage_order: The storage order of the tensor. 0 is row major, and 1 is
                column major.

            strides: Stride along each spatial axis.
        """

        schema = get_schema("MaxPool", 10, "")
        op = Op(self, "MaxPool", schema)
        return op(
            *self._prepare_inputs(schema, X),
            auto_pad=auto_pad,
            ceil_mode=ceil_mode,
            dilations=dilations,
            kernel_shape=kernel_shape,
            pads=pads,
            storage_order=storage_order,
            strides=strides,
        )

    T_Mod = TypeVar(
        "T_Mod",
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Mod(self, A: T_Mod, B: T_Mod, *, fmod: int = 0) -> T_Mod:
        r"""[🌐 Mod(10)](https://onnx.ai/onnx/operators/onnx__Mod.html#mod-10 "Online Documentation")


          Performs element-wise binary modulus (with Numpy-style broadcasting support).
            The sign of the remainder is the same as that of the Divisor.

            Mod operator can also behave like C fmod() or numpy.fmod. In this case, the sign of the remainder however, will be the same as the Dividend
            (in contrast to integer mod). To force a behavior like numpy.fmod() an 'fmod' Attribute is provided.
            This attribute is set to 0 by default causing the behavior to be like integer mod.
            Setting this attribute to 1 causes the remainder to be calculated similar to that of numpy.fmod().

            If the input type is floating point, then `fmod` attribute must be set to 1.

            In case of dividend being zero, the results will be platform dependent.

          This operator supports **multidirectional (i.e., Numpy-style) broadcasting**; for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.


        Args:
            A: Dividend tensor

            B: Divisor tensor

            fmod: Whether the operator should behave like fmod (default=0 meaning it
                will do integer mods); Set this to 1 to force fmod treatment
        """

        schema = get_schema("Mod", 10, "")
        op = Op(self, "Mod", schema)
        return op(*self._prepare_inputs(schema, A, B), fmod=fmod)

    def NonMaxSuppression(
        self,
        boxes: FLOAT,
        scores: FLOAT,
        max_output_boxes_per_class: Optional[INT64] = None,
        iou_threshold: Optional[FLOAT] = None,
        score_threshold: Optional[FLOAT] = None,
        *,
        center_point_box: int = 0,
    ) -> INT64:
        r"""[🌐 NonMaxSuppression(10)](https://onnx.ai/onnx/operators/onnx__NonMaxSuppression.html#nonmaxsuppression-10 "Online Documentation")


        Filter out boxes that have high intersection-over-union (IOU) overlap with previously selected boxes.
        Bounding boxes with score less than score_threshold are removed. Bounding box format is indicated by attribute center_point_box.
        Note that this algorithm is agnostic to where the origin is in the coordinate system and more generally is invariant to
        orthogonal transformations and translations of the coordinate system; thus translating or reflections of the coordinate system
        result in the same boxes being selected by the algorithm.
        The selected_indices output is a set of integers indexing into the input collection of bounding boxes representing the selected boxes.
        The bounding box coordinates corresponding to the selected indices can then be obtained using the Gather or GatherND operation.


        Args:
            boxes: An input tensor with shape [num_batches, spatial_dimension, 4]. The
                single box data format is indicated by center_point_box.

            scores: An input tensor with shape [num_batches, num_classes,
                spatial_dimension]

            max_output_boxes_per_class: (optional) Integer representing the maximum
                number of boxes to be selected per batch per class. It is a scalar.
                Default to 0, which means no output.

            iou_threshold: (optional) Float representing the threshold for deciding
                whether boxes overlap too much with respect to IOU. It is scalar. Value
                range [0, 1]. Default to 0.

            score_threshold: (optional) Float representing the threshold for deciding
                when to remove boxes based on score. It is a scalar.

            center_point_box: Integer indicate the format of the box data. The default
                is 0. 0 - the box data is supplied as [y1, x1, y2, x2] where (y1, x1)
                and (y2, x2) are the coordinates of any diagonal pair of box corners and
                the coordinates can be provided as normalized (i.e., lying in the
                interval [0, 1]) or absolute. Mostly used for TF models. 1 - the box
                data is supplied as [x_center, y_center, width, height]. Mostly used for
                Pytorch models.
        """

        schema = get_schema("NonMaxSuppression", 10, "")
        op = Op(self, "NonMaxSuppression", schema)
        return op(
            *self._prepare_inputs(
                schema,
                boxes,
                scores,
                max_output_boxes_per_class,
                iou_threshold,
                score_threshold,
            ),
            center_point_box=center_point_box,
        )

    T1_QLinearConv = TypeVar("T1_QLinearConv", INT8, UINT8)

    T2_QLinearConv = TypeVar("T2_QLinearConv", INT8, UINT8)

    T3_QLinearConv = TypeVar("T3_QLinearConv", INT8, UINT8)

    T4_QLinearConv: TypeAlias = INT32

    def QLinearConv(
        self,
        x: T1_QLinearConv,
        x_scale: FLOAT,
        x_zero_point: T1_QLinearConv,
        w: T2_QLinearConv,
        w_scale: FLOAT,
        w_zero_point: T2_QLinearConv,
        y_scale: FLOAT,
        y_zero_point: T3_QLinearConv,
        B: Optional[T4_QLinearConv] = None,
        *,
        auto_pad: str = "NOTSET",
        dilations: Optional[Sequence[int]] = None,
        group: int = 1,
        kernel_shape: Optional[Sequence[int]] = None,
        pads: Optional[Sequence[int]] = None,
        strides: Optional[Sequence[int]] = None,
    ) -> T3_QLinearConv:
        r"""[🌐 QLinearConv(10)](https://onnx.ai/onnx/operators/onnx__QLinearConv.html#qlinearconv-10 "Online Documentation")


        The convolution operator consumes a quantized input tensor, its scale and zero point,
        a quantized filter, its scale and zero point, and output's scale and zero point,
        and computes the quantized output. Each scale and zero-point pair must have same shape.
        It means they must be either scalars (per tensor) or 1-D tensors (per output channel).
        Each input or output and its related zero point must have same type.
        When bias is present it must be quantized using scale = input scale * weight scale and
        zero point as 0.


        Args:
            x: Input data tensor from previous layer; has size (N x C x H x W), where N
                is the batch size, C is the number of channels, and H and W are the
                height and width. Note that this is for the 2D image. Otherwise the size
                is (N x C x D1 x D2 ... x Dn). Optionally, if dimension denotation is in
                effect, the operation expects input data tensor to arrive with the
                dimension denotation of [DATA_BATCH, DATA_CHANNEL, DATA_FEATURE,
                DATA_FEATURE ...].

            x_scale: Scale tensor for input 'x'. It's a scalar, which means a
                per-tensor/layer quantization.

            x_zero_point: Zero point tensor for input 'x'. It's a scalar, which means a
                per-tensor/layer quantization.

            w: The weight tensor that will be used in the convolutions; has size (M x
                C/group x kH x kW), where C is the number of channels, and kH and kW are
                the height and width of the kernel, and M is the number of feature maps.
                For more than 2 dimensions, the kernel shape will be (M x C/group x k1 x
                k2 x ... x kn), where (k1 x k2 x ... kn) is the dimension of the kernel.
                Optionally, if dimension denotation is in effect, the operation expects
                the weight tensor to arrive with the dimension denotation of
                [FILTER_OUT_CHANNEL, FILTER_IN_CHANNEL, FILTER_SPATIAL, FILTER_SPATIAL
                ...]. X.shape[1] == (W.shape[1] * group) == C (assuming zero based
                indices for the shape array). Or in other words FILTER_IN_CHANNEL should
                be equal to DATA_CHANNEL.

            w_scale: Scale tensor for input 'w'. It could be a scalar or a 1-D tensor,
                which means a per-tensor/layer or per output channel quantization. If
                it's a 1-D tensor, its number of elements should be equal to the number
                of output channels (M).

            w_zero_point: Zero point tensor for input 'w'. It could be a scalar or a 1-D
                tensor, which means a per-tensor/layer or per output channel
                quantization. If it's a 1-D tensor, its number of elements should be
                equal to the number of output channels (M).

            y_scale: Scale tensor for output 'y'. It's a scalar, which means a
                per-tensor/layer quantization.

            y_zero_point: Zero point tensor for output 'y'. It's a scalar, which means a
                per-tensor/layer quantization.

            B: (optional) Optional 1D bias to be added to the convolution, has size of
                M. Bias must be quantized using scale = x_scale * w_scale and zero_point
                = 0

            auto_pad: auto_pad must be either NOTSET, SAME_UPPER, SAME_LOWER or VALID.
                Where default value is NOTSET, which means explicit padding is used.
                SAME_UPPER or SAME_LOWER mean pad the input so that `output_shape[i] =
                ceil(input_shape[i] / strides[i])` for each axis `i`. The padding is
                split between the two sides equally or almost equally (depending on
                whether it is even or odd). In case the padding is an odd number, the
                extra padding is added at the end for SAME_UPPER and at the beginning
                for SAME_LOWER.

            dilations: dilation value along each spatial axis of the filter. If not
                present, the dilation defaults to 1 along each spatial axis.

            group: number of groups input channels and output channels are divided into.
                default is 1.

            kernel_shape: The shape of the convolution kernel. If not present, should be
                inferred from input 'w'.

            pads: Padding for the beginning and ending along each spatial axis, it can
                take any value greater than or equal to 0.The value represent the number
                of pixels added to the beginning and end part of the corresponding
                axis.`pads` format should be as follow [x1_begin, x2_begin...x1_end,
                x2_end,...], where xi_begin the number ofpixels added at the beginning
                of axis `i` and xi_end, the number of pixels added at the end of axis
                `i`.This attribute cannot be used simultaneously with auto_pad
                attribute. If not present, the padding defaultsto 0 along start and end
                of each spatial axis.

            strides: Stride along each spatial axis. If not present, the stride defaults
                to 1 along each spatial axis.
        """

        schema = get_schema("QLinearConv", 10, "")
        op = Op(self, "QLinearConv", schema)
        return op(
            *self._prepare_inputs(
                schema,
                x,
                x_scale,
                x_zero_point,
                w,
                w_scale,
                w_zero_point,
                y_scale,
                y_zero_point,
                B,
            ),
            auto_pad=auto_pad,
            dilations=dilations,
            group=group,
            kernel_shape=kernel_shape,
            pads=pads,
            strides=strides,
        )

    T1_QLinearMatMul = TypeVar("T1_QLinearMatMul", INT8, UINT8)

    T2_QLinearMatMul = TypeVar("T2_QLinearMatMul", INT8, UINT8)

    T3_QLinearMatMul = TypeVar("T3_QLinearMatMul", INT8, UINT8)

    def QLinearMatMul(
        self,
        a: T1_QLinearMatMul,
        a_scale: FLOAT,
        a_zero_point: T1_QLinearMatMul,
        b: T2_QLinearMatMul,
        b_scale: FLOAT,
        b_zero_point: T2_QLinearMatMul,
        y_scale: FLOAT,
        y_zero_point: T3_QLinearMatMul,
    ) -> T3_QLinearMatMul:
        r"""[🌐 QLinearMatMul(10)](https://onnx.ai/onnx/operators/onnx__QLinearMatMul.html#qlinearmatmul-10 "Online Documentation")


        Matrix product that behaves like [numpy.matmul](https://numpy.org/doc/stable/reference/generated/numpy.matmul.html).
        It consumes two quantized input tensors, their scales and zero points, scale and zero point of output,
        and computes the quantized output. The quantization formula is y = saturate((x / y_scale) + y_zero_point).
        For (x / y_scale), it is rounding to nearest ties to even. Refer to https://en.wikipedia.org/wiki/Rounding for details.
        Scale and zero point must have same shape. They must be either scalar (per tensor) or N-D tensor
        (per row for 'a' and per column for 'b'). Scalar refers to per tensor quantization whereas N-D refers to per row
        or per column quantization. If the input is 2D of shape [M, K] then zero point and scale tensor may be
        an M element vector [v_1, v_2, ..., v_M] for per row quantization and K element vector of shape [v_1, v_2, ..., v_K]
        for per column quantization. If the input is N-D tensor with shape [D1, D2, M, K] then zero point and scale tensor may
        have shape [D1, D2, M, 1] for per row quantization and shape [D1, D2, 1, K] for per column quantization.
        Production must never overflow, and accumulation may overflow if and only if in 32 bits.


        Args:
            a: (non-differentiable) N-dimensional quantized matrix a

            a_scale: (non-differentiable) scale of quantized input a

            a_zero_point: (non-differentiable) zero point of quantized input a

            b: (non-differentiable) N-dimensional quantized matrix b

            b_scale: (non-differentiable) scale of quantized input b

            b_zero_point: (non-differentiable) zero point of quantized input b

            y_scale: (non-differentiable) scale of quantized output y

            y_zero_point: (non-differentiable) zero point of quantized output y
        """

        schema = get_schema("QLinearMatMul", 10, "")
        op = Op(self, "QLinearMatMul", schema)
        return op(
            *self._prepare_inputs(
                schema,
                a,
                a_scale,
                a_zero_point,
                b,
                b_scale,
                b_zero_point,
                y_scale,
                y_zero_point,
            )
        )

    T1_QuantizeLinear = TypeVar("T1_QuantizeLinear", FLOAT, INT32)

    T2_QuantizeLinear = TypeVar("T2_QuantizeLinear", INT8, UINT8)

    def QuantizeLinear(
        self,
        x: T1_QuantizeLinear,
        y_scale: FLOAT,
        y_zero_point: Optional[T2_QuantizeLinear] = None,
    ) -> T2_QuantizeLinear:
        r"""[🌐 QuantizeLinear(10)](https://onnx.ai/onnx/operators/onnx__QuantizeLinear.html#quantizelinear-10 "Online Documentation")


        The linear per-tensor/layer quantization operator. It consumes a high precision tensor, a scale, a zero point to compute the low precision / quantized tensor.
        The quantization formula is y = saturate ((x / y_scale) + y_zero_point). For saturation, it saturates to [0, 255] if it's uint8, or [-128, 127] if it's int8.
        For (x / y_scale), it's rounding to the nearest even. Refer to https://en.wikipedia.org/wiki/Rounding for details. 'y_zero_point' and 'y' must have same type.


        Args:
            x: N-D full precision Input tensor to be quantized.

            y_scale: Scale for doing quantization to get 'y'. It's a scalar, which means
                a per-tensor/layer quantization.

            y_zero_point: (optional) Zero point for doing quantization to get 'y'. It's
                a scalar, which means a per-tensor/layer quantization. Default value is
                uint8 typed 0 if it's not specified.
        """

        schema = get_schema("QuantizeLinear", 10, "")
        op = Op(self, "QuantizeLinear", schema)
        return op(*self._prepare_inputs(schema, x, y_scale, y_zero_point))

    T_Resize = TypeVar(
        "T_Resize",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Resize(self, X: T_Resize, scales: FLOAT, *, mode: str = "nearest") -> T_Resize:
        r"""[🌐 Resize(10)](https://onnx.ai/onnx/operators/onnx__Resize.html#resize-10 "Online Documentation")


        Resize the input tensor.
        Each dimension value of the output tensor is:
          output_dimension = floor(input_dimension * scale).


        Args:
            X: N-D tensor

            scales: The scale array along each dimension. It takes value greater than 0.
                If it's less than 1, it's sampling down, otherwise, it's upsampling. The
                number of elements of 'scales' should be the same as the rank of input
                'X'.

            mode: Two interpolation modes: nearest (default), and linear (including
                bilinear, trilinear, etc)
        """

        schema = get_schema("Resize", 10, "")
        op = Op(self, "Resize", schema)
        return op(*self._prepare_inputs(schema, X, scales), mode=mode)

    T_ReverseSequence = TypeVar(
        "T_ReverseSequence",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def ReverseSequence(
        self,
        input: T_ReverseSequence,
        sequence_lens: INT64,
        *,
        batch_axis: int = 1,
        time_axis: int = 0,
    ) -> T_ReverseSequence:
        r"""[🌐 ReverseSequence(10)](https://onnx.ai/onnx/operators/onnx__ReverseSequence.html#reversesequence-10 "Online Documentation")


        Reverse batch of sequences having different lengths specified by `sequence_lens`.

        For each slice i iterating on batch axis, the operator reverses the first sequence_lens[i] elements on time axis,
        and copies elements whose index's beyond sequence_lens[i] to the output. So the output slice i contains reversed
        sequences on the first sequence_lens[i] elements, then have original values copied for the other elements.

        Example 1:
          input = [[0.0, 4.0, 8.0,  12.0],
                   [1.0, 5.0, 9.0,  13.0],
                   [2.0, 6.0, 10.0, 14.0],
                   [3.0, 7.0, 11.0, 15.0]]
          sequence_lens = [4, 3, 2, 1]
          time_axis = 0
          batch_axis = 1

          output = [[3.0, 6.0, 9.0,  12.0],
                    [2.0, 5.0, 8.0,  13.0],
                    [1.0, 4.0, 10.0, 14.0],
                    [0.0, 7.0, 11.0, 15.0]]

        Example 2:
          input = [[0.0,  1.0,  2.0,  3.0 ],
                   [4.0,  5.0,  6.0,  7.0 ],
                   [8.0,  9.0,  10.0, 11.0],
                   [12.0, 13.0, 14.0, 15.0]]
          sequence_lens = [1, 2, 3, 4]
          time_axis = 1
          batch_axis = 0

          output = [[0.0,  1.0,  2.0,  3.0 ],
                    [5.0,  4.0,  6.0,  7.0 ],
                    [10.0, 9.0,  8.0,  11.0],
                    [15.0, 14.0, 13.0, 12.0]]


        Args:
            input: Tensor of rank r >= 2.

            sequence_lens: Tensor specifying lengths of the sequences in a batch. It has
                shape `[batch_size]`.

            batch_axis: (Optional) Specify which axis is batch axis. Must be one of 1
                (default), or 0.

            time_axis: (Optional) Specify which axis is time axis. Must be one of 0
                (default), or 1.
        """

        schema = get_schema("ReverseSequence", 10, "")
        op = Op(self, "ReverseSequence", schema)
        return op(
            *self._prepare_inputs(schema, input, sequence_lens),
            batch_axis=batch_axis,
            time_axis=time_axis,
        )

    T1_RoiAlign = TypeVar("T1_RoiAlign", DOUBLE, FLOAT, FLOAT16)

    T2_RoiAlign: TypeAlias = INT64

    def RoiAlign(
        self,
        X: T1_RoiAlign,
        rois: T1_RoiAlign,
        batch_indices: T2_RoiAlign,
        *,
        mode: str = "avg",
        output_height: int = 1,
        output_width: int = 1,
        sampling_ratio: int = 0,
        spatial_scale: float = 1.0,
    ) -> T1_RoiAlign:
        r"""[🌐 RoiAlign(10)](https://onnx.ai/onnx/operators/onnx__RoiAlign.html#roialign-10 "Online Documentation")


        Region of Interest (RoI) align operation described in the
        [Mask R-CNN paper](https://arxiv.org/abs/1703.06870).
        RoiAlign consumes an input tensor X and region of interests (rois)
        to apply pooling across each RoI; it produces a 4-D tensor of shape
        (num_rois, C, output_height, output_width).

        RoiAlign is proposed to avoid the misalignment by removing
        quantizations while converting from original image into feature
        map and from feature map into RoI feature; in each ROI bin,
        the value of the sampled locations are computed directly
        through bilinear interpolation.


        Args:
            X: Input data tensor from the previous operator; 4-D feature map of shape
                (N, C, H, W), where N is the batch size, C is the number of channels,
                and H and W are the height and the width of the data.

            rois: RoIs (Regions of Interest) to pool over; rois is 2-D input of shape
                (num_rois, 4) given as [[x1, y1, x2, y2], ...]. The RoIs' coordinates
                are in the coordinate system of the input image. Each coordinate set has
                a 1:1 correspondence with the 'batch_indices' input.

            batch_indices: 1-D tensor of shape (num_rois,) with each element denoting
                the index of the corresponding image in the batch.

            mode: The pooling method. Two modes are supported: 'avg' and 'max'. Default
                is 'avg'.

            output_height: default 1; Pooled output Y's height.

            output_width: default 1; Pooled output Y's width.

            sampling_ratio: Number of sampling points in the interpolation grid used to
                compute the output value of each pooled output bin. If > 0, then exactly
                sampling_ratio x sampling_ratio grid points are used. If == 0, then an
                adaptive number of grid points are used (computed as ceil(roi_width /
                output_width), and likewise for height). Default is 0.

            spatial_scale: Multiplicative spatial scale factor to translate ROI
                coordinates from their input spatial scale to the scale used when
                pooling, i.e., spatial scale of the input feature map X relative to the
                input image. E.g.; default is 1.0f.
        """

        schema = get_schema("RoiAlign", 10, "")
        op = Op(self, "RoiAlign", schema)
        return op(
            *self._prepare_inputs(schema, X, rois, batch_indices),
            mode=mode,
            output_height=output_height,
            output_width=output_width,
            sampling_ratio=sampling_ratio,
            spatial_scale=spatial_scale,
        )

    T_Slice = TypeVar(
        "T_Slice",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    Tind_Slice = TypeVar("Tind_Slice", INT32, INT64)

    def Slice(
        self,
        data: T_Slice,
        starts: Tind_Slice,
        ends: Tind_Slice,
        axes: Optional[Tind_Slice] = None,
        steps: Optional[Tind_Slice] = None,
    ) -> T_Slice:
        r"""[🌐 Slice(10)](https://onnx.ai/onnx/operators/onnx__Slice.html#slice-10 "Online Documentation")


        Produces a slice of the input tensor along multiple axes. Similar to numpy:
        https://numpy.org/doc/stable/reference/routines.indexing.html
        Slices uses `starts`, `ends`, `axes` and `steps` inputs to specify the start and end
        dimension and step for each axis in the list of axes, it uses this information to
        slice the input `data` tensor. If a negative value is passed for any of the
        start or end indices, it represent number of elements before the end of that
        dimension. If the value passed to start or end is larger than the `n` (the
        number of elements in this dimension), it represents `n`. For slicing to the
        end of a dimension with unknown size, it is recommended to pass in `INT_MAX`.
        If a negative value is passed for step, it represents slicing backward.
        If `axes` are omitted, they are set to `[0, ..., ndim-1]`.
        If `steps` are omitted, they are set to `[1, ..., 1]` of length `len(starts)`
        Example 1:
          data = [
              [1, 2, 3, 4],
              [5, 6, 7, 8],
          ]
          axes = [0, 1]
          starts = [1, 0]
          ends = [2, 3]
          steps = [1, 2]
          result = [
              [5, 7],
          ]
        Example 2:
          data = [
              [1, 2, 3, 4],
              [5, 6, 7, 8],
          ]
          starts = [0, 1]
          ends = [-1, 1000]
          result = [
              [2, 3, 4],
          ]


        Args:
            data: Tensor of data to extract slices from.

            starts: 1-D tensor of starting indices of corresponding axis in `axes`

            ends: 1-D tensor of ending indices (exclusive) of corresponding axis in
                `axes`

            axes: (optional) 1-D tensor of axes that `starts` and `ends` apply to.

            steps: (optional) 1-D tensor of slice step of corresponding axis in `axes`.
                Default to 1.
        """

        schema = get_schema("Slice", 10, "")
        op = Op(self, "Slice", schema)
        return op(*self._prepare_inputs(schema, data, starts, ends, axes, steps))

    def StringNormalizer(
        self,
        X: STRING,
        *,
        case_change_action: str = "NONE",
        is_case_sensitive: int = 0,
        locale: Optional[str] = None,
        stopwords: Optional[Sequence[str]] = None,
    ) -> STRING:
        r"""[🌐 StringNormalizer(10)](https://onnx.ai/onnx/operators/onnx__StringNormalizer.html#stringnormalizer-10 "Online Documentation")


        StringNormalization performs string operations for basic cleaning.
        This operator has only one input (denoted by X) and only one output
        (denoted by Y). This operator first examines the elements in the X,
        and removes elements specified in "stopwords" attribute.
        After removing stop words, the intermediate result can be further lowercased,
        uppercased, or just returned depending the "case_change_action" attribute.
        This operator only accepts [C]- and [1, C]-tensor.
        If all elements in X are dropped, the output will be the empty value of string tensor with shape [1]
        if input shape is [C] and shape [1, 1] if input shape is [1, C].


        Args:
            X: UTF-8 strings to normalize

            case_change_action: string enum that cases output to be
                lowercased/uppercases/unchanged. Valid values are "LOWER", "UPPER",
                "NONE". Default is "NONE"

            is_case_sensitive: Boolean. Whether the identification of stop words in X is
                case-sensitive. Default is false

            locale: Environment dependent string that denotes the locale according to
                which output strings needs to be upper/lowercased.Default en_US or
                platform specific equivalent as decided by the implementation.

            stopwords: List of stop words. If not set, no word would be removed from X.
        """

        schema = get_schema("StringNormalizer", 10, "")
        op = Op(self, "StringNormalizer", schema)
        return op(
            *self._prepare_inputs(schema, X),
            case_change_action=case_change_action,
            is_case_sensitive=is_case_sensitive,
            locale=locale,
            stopwords=stopwords,
        )

    T_ThresholdedRelu = TypeVar("T_ThresholdedRelu", DOUBLE, FLOAT, FLOAT16)

    def ThresholdedRelu(
        self, X: T_ThresholdedRelu, *, alpha: float = 1.0
    ) -> T_ThresholdedRelu:
        r"""[🌐 ThresholdedRelu(10)](https://onnx.ai/onnx/operators/onnx__ThresholdedRelu.html#thresholdedrelu-10 "Online Documentation")


        ThresholdedRelu takes one input data (Tensor<T>) and produces one output data
        (Tensor<T>) where the rectified linear function, y = x for x > alpha, y = 0 otherwise,
        is applied to the tensor elementwise.


        Args:
            X: (differentiable) Input tensor

            alpha: Threshold value
        """

        schema = get_schema("ThresholdedRelu", 10, "")
        op = Op(self, "ThresholdedRelu", schema)
        return op(*self._prepare_inputs(schema, X), alpha=alpha)

    T_TopK = TypeVar("T_TopK", DOUBLE, FLOAT, FLOAT16)

    I_TopK: TypeAlias = INT64

    def TopK(self, X: T_TopK, K: INT64, *, axis: int = -1) -> Tuple[T_TopK, I_TopK]:
        r"""[🌐 TopK(10)](https://onnx.ai/onnx/operators/onnx__TopK.html#topk-10 "Online Documentation")


        Retrieve the top-K elements along a specified axis. Given an input tensor of
        shape [a_0, a_1, ..., a_{n-1}] and integer argument k, return two outputs:
          -Value tensor of shape [a_0, a_1, ..., a_{axis-1}, k, a_{axis+1}, ... a_{n-1}]
            which contains the values of the top k elements along the specified axis
          -Index tensor of shape [a_0, a_1, ..., a_{axis-1}, k, a_{axis+1}, ... a_{n-1}] which
           contains the indices of the top k elements (original indices from the input
           tensor).

        Given two equivalent values, this operator uses the indices along the axis  as
         a tiebreaker. That is, the element with the lower index will appear first.


        Args:
            X: Tensor of shape [a_0, a_1, ..., a_{n-1}]

            K: A 1-D tensor containing a single positive value corresponding to the
                number of top elements to retrieve

            axis: Dimension on which to do the sort.
        """

        schema = get_schema("TopK", 10, "")
        op = Op(self, "TopK", schema)
        return op(*self._prepare_inputs(schema, X, K), axis=axis)



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

"""Typings for function definitions."""

from __future__ import annotations

from typing import TypeVar, Union

from onnxscript import (
    BFLOAT16,
    BOOL,
    COMPLEX64,
    COMPLEX128,
    DOUBLE,
    FLOAT,
    FLOAT16,
    INT8,
    INT16,
    INT32,
    INT64,
    STRING,
    UINT8,
)

# NOTE: We do not care about unsigned types beyond UINT8 because PyTorch does not us them.
# More detail can be found: https://pytorch.org/docs/stable/tensors.html

_TensorType = Union[
    BFLOAT16,
    BOOL,
    COMPLEX64,
    COMPLEX128,
    DOUBLE,
    FLOAT,
    FLOAT16,
    INT8,
    INT16,
    INT32,
    INT64,
    UINT8,
]
_FloatType = Union[FLOAT16, FLOAT, DOUBLE, BFLOAT16]
IntType = Union[INT8, INT16, INT32, INT64]
RealType = Union[
    BFLOAT16,
    FLOAT16,
    FLOAT,
    DOUBLE,
    INT8,
    INT16,
    INT32,
    INT64,
]

TTensor = TypeVar("TTensor", bound=_TensorType)
# Duplicate TTensor for inputs/outputs that accept the same set of types as TTensor
# but do not constrain the type to be the same as the other inputs/outputs
TTensor2 = TypeVar("TTensor2", bound=_TensorType)
TTensorOrString = TypeVar("TTensorOrString", bound=Union[_TensorType, STRING])
TFloat = TypeVar("TFloat", bound=_FloatType)
TFloatOrUInt8 = TypeVar("TFloatOrUInt8", bound=Union[FLOAT, FLOAT16, DOUBLE, INT8, UINT8])
TInt = TypeVar("TInt", bound=IntType)
TReal = TypeVar("TReal", bound=RealType)
TRealUnlessInt16OrInt8 = TypeVar(
    "TRealUnlessInt16OrInt8", bound=Union[FLOAT16, FLOAT, DOUBLE, BFLOAT16, INT32, INT64]
)
TRealUnlessFloat16OrInt8 = TypeVar(
    "TRealUnlessFloat16OrInt8", bound=Union[DOUBLE, FLOAT, INT16, INT32, INT64]
)
TRealOrUInt8 = TypeVar("TRealOrUInt8", bound=Union[RealType, UINT8])
TFloatHighPrecision = TypeVar("TFloatHighPrecision", bound=Union[FLOAT, DOUBLE])



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""SDPA fusion test cases."""

from __future__ import annotations

import math
import unittest

import numpy
import onnx_ir as ir
import parameterized

import onnxscript.optimizer
from onnxscript import script
from onnxscript.onnx_opset import opset18 as op
from onnxscript.onnx_types import FLOAT
from onnxscript.rewriter.ort_fusions._test_utils import assert_allclose, ort_run
from onnxscript.rewriter.ort_fusions.sdpa import fuse_sdpa
from onnxscript.rewriter.ort_fusions.sdpa_via_mha import replace_sdpa_by_mha

B = 2  # batch size
N = 4  # number of heads
S = 8  # sequence length
H = 128  # head size
SCALE_FACTOR = math.sqrt(H)
MUL_SCALE_FACTOR = 1.0 / SCALE_FACTOR
SQRT_SCALE_FACTOR = math.sqrt(SCALE_FACTOR)
SQRT_MUL_SCALE_FACTOR = math.sqrt(MUL_SCALE_FACTOR)
# Custom scale factors for testing
CUSTOM_SCALE_FACTOR = 1.0 / math.sqrt(80)
CUSTOM_MUL_SCALE_FACTOR = CUSTOM_SCALE_FACTOR
CUSTOM_DIV_SCALE_FACTOR = 1.0 / CUSTOM_SCALE_FACTOR
SQRT_CUSTOM_MUL_SCALE_FACTOR = math.sqrt(CUSTOM_MUL_SCALE_FACTOR)
SQRT_CUSTOM_DIV_SCALE_FACTOR = math.sqrt(CUSTOM_DIV_SCALE_FACTOR)


@script()
def _unmasked_pre_div_sdpa_script(query, key, value):
    key_transposed = op.Transpose(key, perm=[0, 1, 3, 2])
    divisor = op.Constant(value_float=SQRT_SCALE_FACTOR)
    scaled_query = op.Div(query, divisor)
    scaled_key = op.Div(key_transposed, divisor)
    attn_score = op.MatMul(scaled_query, scaled_key)
    attn_weight = op.Softmax(attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


@script()
def _unmasked_pre_mul_sdpa_script(query, key, value):
    key_transposed = op.Transpose(key, perm=[0, 1, 3, 2])
    multiplier = op.Constant(value_float=SQRT_MUL_SCALE_FACTOR)
    scaled_query = op.Mul(query, multiplier)
    scaled_key = op.Mul(key_transposed, multiplier)
    attn_score = op.MatMul(scaled_query, scaled_key)
    attn_weight = op.Softmax(attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


@script()
def _unmasked_post_div_sdpa_script(query, key, value):
    key_transposed = op.Transpose(key, perm=[0, 1, 3, 2])
    divisor = op.Constant(value_float=SCALE_FACTOR)
    attn_score = op.MatMul(query, key_transposed)
    scaled_attn_score = op.Div(attn_score, divisor)
    attn_weight = op.Softmax(scaled_attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


@script()
def _unmasked_post_mul_sdpa_script(query, key, value):
    key_transposed = op.Transpose(key, perm=[0, 1, 3, 2])
    multiplier = op.Constant(value_float=MUL_SCALE_FACTOR)
    attn_score = op.MatMul(query, key_transposed)
    scaled_attn_score = op.Mul(attn_score, multiplier)
    attn_weight = op.Softmax(scaled_attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


@script()
def _custom_scale_pre_div_sdpa_script(query, key, value):
    key_transposed = op.Transpose(key, perm=[0, 1, 3, 2])
    divisor = op.Constant(value_float=SQRT_CUSTOM_DIV_SCALE_FACTOR)
    scaled_query = op.Div(query, divisor)
    scaled_key = op.Div(key_transposed, divisor)
    attn_score = op.MatMul(scaled_query, scaled_key)
    attn_weight = op.Softmax(attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


@script()
def _custom_scale_pre_mul_sdpa_script(query, key, value):
    key_transposed = op.Transpose(key, perm=[0, 1, 3, 2])
    multiplier = op.Constant(value_float=SQRT_CUSTOM_MUL_SCALE_FACTOR)
    scaled_query = op.Mul(query, multiplier)
    scaled_key = op.Mul(key_transposed, multiplier)
    attn_score = op.MatMul(scaled_query, scaled_key)
    attn_weight = op.Softmax(attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


@script()
def _custom_multi_scale_pre_mul_sdpa_script(query, key, value):
    key_transposed = op.Transpose(key, perm=[0, 1, 3, 2])
    multiplier_q = op.Constant(value_float=SQRT_CUSTOM_MUL_SCALE_FACTOR)
    multiplier_k = op.Constant(value_float=SQRT_CUSTOM_MUL_SCALE_FACTOR)
    scaled_query = op.Mul(query, multiplier_q)
    scaled_key = op.Mul(key_transposed, multiplier_k)
    attn_score = op.MatMul(scaled_query, scaled_key)
    attn_weight = op.Softmax(attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


@script()
def _custom_scale_post_div_sdpa_script(query, key, value):
    key_transposed = op.Transpose(key, perm=[0, 1, 3, 2])
    divisor = op.Constant(value_float=CUSTOM_DIV_SCALE_FACTOR)
    attn_score = op.MatMul(query, key_transposed)
    scaled_attn_score = op.Div(attn_score, divisor)
    attn_weight = op.Softmax(scaled_attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


@script()
def _custom_scale_post_mul_sdpa_script(query, key, value):
    key_transposed = op.Transpose(key, perm=[0, 1, 3, 2])
    multiplier = op.Constant(value_float=CUSTOM_MUL_SCALE_FACTOR)
    attn_score = op.MatMul(query, key_transposed)
    scaled_attn_score = op.Mul(attn_score, multiplier)
    attn_weight = op.Softmax(scaled_attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


@script()
def _masked_pre_div_sdpa_script(query, key, value, mask):
    key_transposed = op.Transpose(key, perm=[0, 1, 3, 2])
    divisor = op.Constant(value_float=SQRT_SCALE_FACTOR)
    scaled_query = op.Div(query, divisor)
    scaled_key = op.Div(key_transposed, divisor)
    attn_score = op.MatMul(scaled_query, scaled_key)
    masked_attn_score = op.Add(attn_score, mask)
    attn_weight = op.Softmax(masked_attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


@script()
def _masked_pre_mul_sdpa_script(query, key, value, mask):
    key_transposed = op.Transpose(key, perm=[0, 1, 3, 2])
    multiplier = op.Constant(value_float=SQRT_MUL_SCALE_FACTOR)
    scaled_query = op.Mul(query, multiplier)
    scaled_key = op.Mul(key_transposed, multiplier)
    attn_score = op.MatMul(scaled_query, scaled_key)
    masked_attn_score = op.Add(attn_score, mask)
    attn_weight = op.Softmax(masked_attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


@script()
def _masked_post_div_sdpa_script(query, key, value, mask):
    key_transposed = op.Transpose(key, perm=[0, 1, 3, 2])
    divisor = op.Constant(value_float=SCALE_FACTOR)
    attn_score = op.MatMul(query, key_transposed)
    scaled_attn_score = op.Div(attn_score, divisor)
    masked_attn_score = op.Add(scaled_attn_score, mask)
    attn_weight = op.Softmax(masked_attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


@script()
def _masked_post_mul_sdpa_script(query, key, value, mask):
    key_transposed = op.Transpose(key, perm=[0, 1, 3, 2])
    multiplier = op.Constant(value_float=MUL_SCALE_FACTOR)
    attn_score = op.MatMul(query, key_transposed)
    scaled_attn_score = op.Mul(attn_score, multiplier)
    masked_attn_score = op.Add(scaled_attn_score, mask)
    attn_weight = op.Softmax(masked_attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


@script()
def _masked_custom_scale_pre_div_sdpa_script(query, key, value, mask):
    key_transposed = op.Transpose(key, perm=[0, 1, 3, 2])
    divisor = op.Constant(value_float=SQRT_CUSTOM_DIV_SCALE_FACTOR)
    scaled_query = op.Div(query, divisor)
    scaled_key = op.Div(key_transposed, divisor)
    attn_score = op.MatMul(scaled_query, scaled_key)
    masked_attn_score = op.Add(attn_score, mask)
    attn_weight = op.Softmax(masked_attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


@script()
def _masked_custom_scale_pre_mul_sdpa_script(query, key, value, mask):
    key_transposed = op.Transpose(key, perm=[0, 1, 3, 2])
    multiplier = op.Constant(value_float=SQRT_CUSTOM_MUL_SCALE_FACTOR)
    scaled_query = op.Mul(query, multiplier)
    scaled_key = op.Mul(key_transposed, multiplier)
    attn_score = op.MatMul(scaled_query, scaled_key)
    masked_attn_score = op.Add(attn_score, mask)
    attn_weight = op.Softmax(masked_attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


@script()
def _masked_custom_scale_post_div_sdpa_script(query, key, value, mask):
    key_transposed = op.Transpose(key, perm=[0, 1, 3, 2])
    divisor = op.Constant(value_float=CUSTOM_DIV_SCALE_FACTOR)
    attn_score = op.MatMul(query, key_transposed)
    scaled_attn_score = op.Div(attn_score, divisor)
    masked_attn_score = op.Add(scaled_attn_score, mask)
    attn_weight = op.Softmax(masked_attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


@script()
def _masked_custom_scale_post_mul_sdpa_script(query, key, value, mask):
    key_transposed = op.Transpose(key, perm=[0, 1, 3, 2])
    multiplier = op.Constant(value_float=CUSTOM_MUL_SCALE_FACTOR)
    attn_score = op.MatMul(query, key_transposed)
    scaled_attn_score = op.Mul(attn_score, multiplier)
    masked_attn_score = op.Add(scaled_attn_score, mask)
    attn_weight = op.Softmax(masked_attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


# This tests a scenario where the key is in BSHd format instead of BHSd, which
# happens due to an optimization that fuses two transposes together, the one
# to convert from BSHd to BHSd and then to BHdS before MatMul. Hence, the first
# transpose down below is different from other test cases.
@script()
def _unmasked_pre_div_sdpa_BSHd_key_script(query, key, value):
    key_transposed = op.Transpose(key, perm=[0, 2, 3, 1])  # BSHd to BHdS
    divisor = op.Constant(value_float=SQRT_SCALE_FACTOR)
    scaled_query = op.Div(query, divisor)
    scaled_key = op.Div(key_transposed, divisor)
    attn_score = op.MatMul(scaled_query, scaled_key)
    attn_weight = op.Softmax(attn_score, axis=-1)
    is_nan = op.IsNaN(attn_weight)
    zero = op.Constant(value_float=0.0)
    adj_attn_weight = op.Where(is_nan, zero, attn_weight)
    attn_output = op.MatMul(adj_attn_weight, value)
    return attn_output


class SDPATestCase:
    def __init__(self, script_func, *, with_mask, BSHd_key=False):
        self.script_func = script_func
        self.with_mask = with_mask
        self.BSHd_key = BSHd_key

    def get_onnx_model(self):
        if not hasattr(self, "_onnx_model"):
            qv_type = FLOAT[B, N, S, H]
            mask_type = FLOAT[B, N, S, S]
            k_type = FLOAT[B, S, N, H] if self.BSHd_key else FLOAT[B, N, S, H]
            input_types = [qv_type, k_type, qv_type]
            if self.with_mask:
                input_types.append(mask_type)
            model_proto = self.script_func.to_model_proto(
                input_types=input_types, output_types=[qv_type]
            )
            self._onnx_model = ir.serde.deserialize_model(model_proto)
        return self._onnx_model

    def get_ort_inputs(self):
        if not hasattr(self, "_ort_inputs"):
            inputs = {
                "query": numpy.random.rand(B, N, S, H).astype(numpy.float32),
                "key": numpy.random.rand(B, S, N, H).astype(numpy.float32)
                if self.BSHd_key
                else numpy.random.rand(B, N, S, H).astype(numpy.float32),
                "value": numpy.random.rand(B, N, S, H).astype(numpy.float32),
            }
            if self.with_mask:
                inputs["mask"] = numpy.random.rand(B, N, S, S).astype(numpy.float32)
            self._ort_inputs = inputs
        return self._ort_inputs


class InvalidSDPATestCase:
    def __init__(self, script_func):
        self.script_func = script_func

    def get_onnx_model(self):
        if not hasattr(self, "_onnx_model"):
            qk_type = FLOAT[B, N, S, H]
            # We broadcast value in the batch dimension, which is not supported by SDPA fusion
            v_type = FLOAT[1, N, S, H]
            mask_type = FLOAT[B, N, S, S]
            model_proto = self.script_func.to_model_proto(
                input_types=[qk_type, qk_type, v_type, mask_type], output_types=[qk_type]
            )
            self._onnx_model = ir.serde.deserialize_model(model_proto)
        return self._onnx_model

    def get_ort_inputs(self):
        if not hasattr(self, "_ort_inputs"):
            inputs = {
                "query": numpy.random.rand(B, N, S, H).astype(numpy.float32),
                "key": numpy.random.rand(B, N, S, H).astype(numpy.float32),
                "value": numpy.random.rand(1, N, S, H).astype(numpy.float32),
                "mask": numpy.random.rand(B, N, S, S).astype(numpy.float32),
            }
            self._ort_inputs = inputs
        return self._ort_inputs


class TestSDPAFusion(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            ("pre_div", _unmasked_pre_div_sdpa_script),
            ("pre_mul", _unmasked_pre_mul_sdpa_script),
            ("post_div", _unmasked_post_div_sdpa_script),
            ("post_mul", _unmasked_post_mul_sdpa_script),
            ("masked_pre_div", _masked_pre_div_sdpa_script),
            ("masked_pre_mul", _masked_pre_mul_sdpa_script),
            ("masked_post_div", _masked_post_div_sdpa_script),
            ("masked_post_mul", _masked_post_mul_sdpa_script),
            ("custom_scale_post_mul", _custom_scale_post_mul_sdpa_script),
            ("custom_scale_post_div", _custom_scale_post_div_sdpa_script),
            ("custom_scale_pre_mul", _custom_scale_pre_mul_sdpa_script),
            ("custom_scale_pre_div", _custom_scale_pre_div_sdpa_script),
            ("masked_custom_scale_post_mul", _masked_custom_scale_post_mul_sdpa_script),
            ("masked_custom_scale_post_div", _masked_custom_scale_post_div_sdpa_script),
            ("masked_custom_scale_pre_mul", _masked_custom_scale_pre_mul_sdpa_script),
            ("masked_custom_scale_pre_div", _masked_custom_scale_pre_div_sdpa_script),
            (
                "_custom_multi_scale_pre_mul_sdpa_script",
                _custom_multi_scale_pre_mul_sdpa_script,
            ),
            ("pre_div_sdpa_BSHd_key", _unmasked_pre_div_sdpa_BSHd_key_script),
        ]
    )
    def test_sdpa_fusion(self, name, script_func):
        test_case = SDPATestCase(
            script_func, with_mask="masked" in name, BSHd_key="BSHd_key" in name
        )
        model = test_case.get_onnx_model()
        onnxscript.optimizer.optimize(model)

        inputs = test_case.get_ort_inputs()
        original_outputs = ort_run("original", model, inputs)

        count = fuse_sdpa(model, debug=True)
        self.assertGreater(count, 0)

        # Check that the fusion was successful
        op_types = [n.op_type for n in model.graph]
        self.assertIn("SDPA", op_types)

        # Ensure that the scale of the SDPA node is set correctly
        sdpa_node = next(n for n in model.graph if n.op_type == "SDPA")
        self.assertEqual(sdpa_node.op_type, "SDPA")

        if "custom" in name:
            self.assertIsNotNone(sdpa_node.attributes.get("scale"))
            scale_factor = sdpa_node.attributes["scale"].value
            self.assertAlmostEqual(scale_factor, CUSTOM_SCALE_FACTOR, delta=1e-8)
        else:
            # These tests are for the default scaling factors, no scale factor is passed to SDPA
            # pattern rewriting check functions should be sufficient to check if expected value
            # of scale_factor (is =default_scaling_factor)
            self.assertIsNone(sdpa_node.attributes.get("scale"))

        replace_sdpa_by_mha(model, debug=True)

        self.assertNotIn("SDPA", [n.op_type for n in model.graph])

        new_outputs = ort_run("optimized", model, inputs)
        assert_allclose(new_outputs, original_outputs)

    def test_invalid_sdpa_fusion_value_batch_dim(self):
        test_case = InvalidSDPATestCase(_masked_pre_mul_sdpa_script)
        model = test_case.get_onnx_model()
        onnxscript.optimizer.optimize(model)
        count = fuse_sdpa(model)
        self.assertEqual(count, 0)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

from typing import Sequence, Union

import onnx_ir as ir

from onnxscript.rewriter import _fusion_utils, _ir_utils, pattern

Dim = Union[int, ir.SymbolicDim]


# TODO: Maybe add this check to utilities


class AttentionFusion(pattern.RewriteRuleClassBase):
    def __init__(
        self,
        name,
        *,
        has_past: bool,
        no_slice: bool,
    ):
        super().__init__(name)
        self._has_past = has_past
        self._no_slice = no_slice

    def pattern(
        self,
        op,
        input,
        qkv_weight,
        qkv_bias,
        # mask_index,
        past,
        num_heads,
        # scale,
        start1,
        end1,
        start2,
        end2,
        start3,
        end3,
        q_mul,
        k_mul,
        v_mul,
    ):
        if self._no_slice:
            query_BSD = op.MatMul(input, q_mul)
            key_BSD = op.MatMul(input, k_mul)
            value_BSD = op.MatMul(input, v_mul)
        else:
            projected = op.MatMul(input, qkv_weight, _outputs=["projected"])

            # Slice packed Matmul QKV into Q, K, and V
            # Q, K, and V are of shape (B, S, D)
            query_BSD = op.Slice(
                projected,
                start1,  # starts
                end1,  # ends
                [2],  # axes
                _outputs=["query_mm_sliced"],
            )
            key_BSD = op.Slice(
                projected,
                start2,  # starts
                end2,  # ends
                [2],  # axes
                _outputs=["key_mm_sliced"],
            )
            value_BSD = op.Slice(
                projected,
                start3,  # starts
                end3,  # ends
                [2],  # axes
                _outputs=["value_mm_sliced"],
            )

        # TODO: Add other attributes

        if self._has_past:
            # Split past into past_key and past_value
            # past_key and past_value are of shape (B, H, S, D/H)
            past_key = op.Slice(
                past,
                [0],  # starts
                [1],  # ends
                [0],  # axes
                _outputs=["past_key_sliced"],
            )
            past_key = op.Squeeze(past_key, [0])
            past_value = op.Slice(
                past,
                [1],  # starts
                [2],  # ends
                [0],  # axes
                _outputs=["past_value_sliced"],
            )
            past_value = op.Squeeze(past_value, [0])

            attention, present_key, present_value = op.MultiHeadAttention(
                query_BSD,
                key_BSD,
                value_BSD,
                qkv_bias,
                None,  # key_padding_mask
                pattern.Var("attention_bias", can_match_none=True),
                past_key,
                past_value,
                num_heads=num_heads,
                # scale=scale,
                _domain="com.microsoft",
                _outputs=["mha_output", "present_key", "present_value"],
            )
            # Concat present_key and present_value to form present
            present_key = op.Unsqueeze(present_key, [0])
            present_value = op.Unsqueeze(present_value, [0])
            present = op.Concat(present_key, present_value, axis=0)
            # Return present output first as it captures the complete pattern graph
            return present, attention
        else:
            attention = op.MultiHeadAttention(
                query_BSD,
                key_BSD,
                value_BSD,
                qkv_bias,
                None,  # key_padding_mask
                pattern.Var("attention_bias", can_match_none=True),
                None,  # past_key
                None,  # past_value
                num_heads=num_heads,
                # scale=scale,
                _domain="com.microsoft",
                _outputs=["mha_output"],
            )
            return attention

    def check(
        self,
        op,
        input,
        qkv_weight,
        projected=None,
        query_mm_sliced=None,
        key_mm_sliced=None,
        value_mm_sliced=None,
        start1=None,
        end1=None,
        start2=None,
        end2=None,
        start3=None,
        end3=None,
        q_mul=None,
        k_mul=None,
        v_mul=None,
        **_,
    ):
        check_result = pattern.MatchResult()
        self.bindings: dict[str, Dim] = {}

        def no_match(val: ir.Value, dims: Sequence[str]) -> bool:
            return not _fusion_utils.check_shape_bool(self.bindings, val, dims)

        if no_match(input, ["B", "S", "D"]):
            return check_result.fail(
                f"Shape mismatch: {input} does not match expected dimensions ['B', 'S', 'D']",
                input,
            )
        if not self._no_slice:
            # Ensure slicing is done correctly
            if projected is None or projected.shape is None or len(projected.shape) != 3:
                return check_result.fail("Input projection is not a 3D tensor.", projected)
            hidden_size = projected.shape[2]
            if not isinstance(hidden_size, int):
                return check_result.fail("Hidden size is not an integer.", projected)
            if not (
                _ir_utils.is_singleton_value(start1, 0)
                and _ir_utils.get_singleton_value(end1)
                == _ir_utils.get_singleton_value(start2)
                and _ir_utils.get_singleton_value(end2)
                == _ir_utils.get_singleton_value(start3)
                and _ir_utils.is_singleton_value(end3, lambda x: x >= hidden_size)
            ):
                return check_result.fail(
                    "Projected input is not being split into q, k, v correctly based on hidden sizes.",
                    projected,
                )

            if no_match(qkv_weight, ["D", "Dh"]):
                return check_result.fail(
                    f"Shape mismatch: {qkv_weight} does not match expected dimensions ['D', 'Dh']",
                    qkv_weight,
                )
            if no_match(query_mm_sliced, ["B", "S", "Dh_q"]):
                return check_result.fail(
                    f"Shape mismatch: {query_mm_sliced} does not match expected dimensions ['B', 'S', 'Dh_q']",
                    query_mm_sliced,
                )
            if no_match(key_mm_sliced, ["B", "S", "Dh_k"]):
                return check_result.fail(
                    f"Shape mismatch: {key_mm_sliced} does not match expected dimensions ['B', 'S', 'Dh_k']",
                    key_mm_sliced,
                )
            if no_match(value_mm_sliced, ["B", "S", "Dh_v"]):
                return check_result.fail(
                    f"Shape mismatch: {value_mm_sliced} does not match expected dimensions ['B', 'S', 'Dh_v']",
                    value_mm_sliced,
                )
        else:
            if no_match(q_mul, ["D", "Dh_q"]):
                return check_result.fail(
                    f"Shape mismatch: {q_mul} does not match expected dimensions ['D', 'Dh_q']",
                    q_mul,
                )
            if no_match(k_mul, ["D", "Dh_k"]):
                return check_result.fail(
                    f"Shape mismatch: {k_mul} does not match expected dimensions ['D', 'Dh_k']",
                    k_mul,
                )
            if no_match(v_mul, ["D", "Dh_v"]):
                return check_result.fail(
                    f"Shape mismatch: {v_mul} does not match expected dimensions ['D', 'Dh_v']",
                    v_mul,
                )

        # Ensure Dh = Dh_q + Dh_k + Dh_v
        Dh = self.bindings.get("Dh")
        Dh_q = self.bindings.get("Dh_q")
        Dh_k = self.bindings.get("Dh_k")
        Dh_v = self.bindings.get("Dh_v")

        if not isinstance(Dh_q, int) or not isinstance(Dh_k, int) or not isinstance(Dh_v, int):
            return check_result.fail(
                "Could not determine the hidden sizes of query, key, and value.",
            )

        if not self._no_slice:
            if not isinstance(Dh, int):
                return check_result.fail(
                    "Could not determine the total hidden size of weight.",
                )

            if Dh != Dh_q + Dh_k + Dh_v:  # type: ignore[operator]
                return check_result.fail(
                    f"Hidden size of query, key and value do not add up to hidden size: {Dh} != {Dh_q} + {Dh_k} + {Dh_v}",
                )

        # TODO: Add mask check once mask is added to the pattern
        return check_result

    def rewrite(
        self,
        op,
        input,
        qkv_weight,
        qkv_bias,
        # mask_index,
        past,
        attention_bias,
        num_heads,
        # scale,
        mha_output,
        q_mul=None,
        k_mul=None,
        v_mul=None,
        **_,
    ):
        # Use bindings to get the values of Dh_q, Dh_k, and Dh_v
        # and construct qkv_hidden_sizes
        Dh_q = self.bindings.get("Dh_q")
        Dh_k = self.bindings.get("Dh_k")
        Dh_v = self.bindings.get("Dh_v")
        qkv_hidden_sizes = [Dh_q, Dh_k, Dh_v]
        if self._no_slice:
            qkv_weight = op.Concat(q_mul, k_mul, v_mul, axis=1)

        scale = mha_output.producer().attributes.get_float("scale", None)

        if self._has_past:
            attention, present = op.Attention(
                input,
                qkv_weight,
                qkv_bias,
                None,  # mask_index
                past,
                attention_bias,
                # past_sequence_length
                num_heads=num_heads,
                qkv_hidden_sizes=qkv_hidden_sizes,
                scale=scale,
                _domain="com.microsoft",
                _outputs=2,
            )
            # Use same output ordering as in pattern
            return present, attention
        else:
            return op.Attention(
                input,
                qkv_weight,
                qkv_bias,
                None,  # mask_index
                None,  # past
                attention_bias,
                None,  # past_sequence_length
                num_heads=num_heads,
                qkv_hidden_sizes=qkv_hidden_sizes,
                scale=scale,
                _domain="com.microsoft",
                _outputs=1,
            )


# Define all combinations of parameters
parameter_combinations = [
    {
        "name": f"attention_{'with_past_' if has_past else ''}{'no_slice' if no_slice else ''}".strip(
            "_"
        ),
        "has_past": has_past,
        "no_slice": no_slice,
    }
    for has_past in [False, True]
    for no_slice in [False, True]
]

# Dynamically create the rules
attention_rules = pattern.RewriteRuleSet(
    [
        AttentionFusion.rule(
            params["name"],
            has_past=params["has_past"],
            no_slice=params["no_slice"],
        )
        for params in parameter_combinations
    ]
)


fuse_attention = _fusion_utils.apply_fusion_rules(attention_rules)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Small test case models for rotary embedding."""

import numpy
import onnx_ir as ir

from onnxscript import script
from onnxscript.onnx_opset import opset18 as op
from onnxscript.onnx_types import FLOAT, INT64

# A simple rotary embedding example


# x: [B, H, S, E]
# position_ids: [B, S]
@script()
def _test_case_1_script(x: FLOAT[1, 4, 8, 8], position_ids: INT64[1, 8]) -> FLOAT[1, 4, 8, 8]:
    inv_freq = op.Constant(value_floats=[1.0, 2.0, 3.0, 4.0])
    inv_freq_3d = op.Unsqueeze(inv_freq, [0, 2])
    position_ids_expanded = op.Unsqueeze(position_ids, [1])  # => [B, 1, S]
    position_ids_float = op.Cast(position_ids_expanded, to=ir.DataType.FLOAT)
    freqs = op.MatMul(inv_freq_3d, position_ids_float)  # [B, E, S]
    freqs = op.Transpose(freqs, perm=[0, 2, 1])  # [B, S, E]
    emb = op.Concat(freqs, freqs, axis=-1)
    cos = op.Cos(emb)
    sin = op.Sin(emb)
    cos_4d = op.Unsqueeze(cos, [1])
    sin_4d = op.Unsqueeze(sin, [1])

    x1 = op.Slice(x, [0], [4], [3], [1])
    x2 = op.Slice(x, [4], [8], [3], [1])
    minus_x2 = op.Neg(x2)
    rotated_x = op.Concat(minus_x2, x1, axis=-1)
    rotary_embedding = op.Add(x * cos_4d, rotated_x * sin_4d)
    return rotary_embedding


class _TestCase1:
    def get_onnx_model(self):
        if not hasattr(self, "_onnx_model"):
            model_proto = _test_case_1_script.to_model_proto()
            model = ir.serde.deserialize_model(model_proto)
            self._onnx_model = model
        return self._onnx_model

    def get_ort_inputs(self):
        if not hasattr(self, "_ort_inputs"):
            inputs = {
                "x": numpy.random.rand(1, 4, 8, 8).astype(numpy.float32),
                "position_ids": numpy.arange(8, dtype=numpy.int64).reshape(1, 8),
            }
            self._ort_inputs = inputs
        return self._ort_inputs


def test_case_1():
    return _TestCase1()


# A simple rotary embedding example with 1D position_ids
# x: [B, H, S, E]
# position_ids: [S]
@script()
def _test_case_2_script(x: FLOAT[1, 4, 8, 8], position_ids: INT64[8]) -> FLOAT[1, 4, 8, 8]:
    inv_freq = op.Constant(value_floats=[1.0, 2.0, 3.0, 4.0])
    inv_freq_3d = op.Unsqueeze(inv_freq, [0, 2])
    position_ids_expanded = op.Unsqueeze(position_ids, [0, 1])  # => [1, 1, S]
    position_ids_float = op.Cast(position_ids_expanded, to=ir.DataType.FLOAT)
    freqs = op.MatMul(inv_freq_3d, position_ids_float)  # [B, E, S]
    freqs = op.Transpose(freqs, perm=[0, 2, 1])  # [B, S, E]
    emb = op.Concat(freqs, freqs, axis=-1)
    cos = op.Cos(emb)
    sin = op.Sin(emb)
    cos_4d = op.Unsqueeze(cos, [1])
    sin_4d = op.Unsqueeze(sin, [1])

    x1 = op.Slice(x, [0], [4], [3], [1])
    x2 = op.Slice(x, [4], [8], [3], [1])
    minus_x2 = op.Neg(x2)
    rotated_x = op.Concat(minus_x2, x1, axis=-1)
    rotary_embedding = op.Add(x * cos_4d, rotated_x * sin_4d)
    return rotary_embedding


class _TestCase2:
    def get_onnx_model(self):
        if not hasattr(self, "_onnx_model"):
            model_proto = _test_case_2_script.to_model_proto()
            model = ir.serde.deserialize_model(model_proto)
            self._onnx_model = model
        return self._onnx_model

    def get_ort_inputs(self):
        if not hasattr(self, "_ort_inputs"):
            inputs = {
                "x": numpy.random.rand(1, 4, 8, 8).astype(numpy.float32),
                "position_ids": numpy.arange(8, dtype=numpy.int64).reshape(8),
            }
            self._ort_inputs = inputs
        return self._ort_inputs


def test_case_2():
    return _TestCase2()


# A partial rotary embedding example:

rotary_embedding_dim = 32  # Abbreviated as "rd" in shape descriptors below
half_rotary_embedding_dim = rotary_embedding_dim // 2
# A random inverse frequency tensor for the sake of this example.
inv_freqs_value = numpy.random.rand(1, half_rotary_embedding_dim, 1).astype(numpy.float32)


@script()
def _partial_rotary_script(position_ids, query):
    inv_freqs = op.Constant(value=inv_freqs_value)  # [1, rd/2, 1]
    position_ids_3d = op.Unsqueeze(position_ids, 1)  # [B, 1, S]
    position_ids_3d_float = op.Cast(position_ids_3d, to=1)
    matmul = op.MatMul(inv_freqs, position_ids_3d_float)  # [B, rd/2, S]
    transpose = op.Transpose(matmul, perm=[0, 2, 1])  # [B, S, rd/2]
    cat = op.Concat(transpose, transpose, axis=-1)  # [B, S, rd]
    cos_3d = op.Cos(cat)  # [B, S, rd]
    sin_3d = op.Sin(cat)  # [B, S, rd]
    # Split the query for partial embedding
    to_embed = op.Slice(query, [0], [32], [3], [1])
    unembedded = op.Slice(query, [32], [9223372036854775807], [3], [1])
    cos_4d = op.Unsqueeze(cos_3d, [1])  # [B, 1, S, rd]
    sin_4d = op.Unsqueeze(sin_3d, [1])  # [B, 1, S, rd]
    # Compute rotation of X as X * cos + rotate_half(X) * sin, where rotate_half(X)
    # essentially represents X rotated by 90 degrees
    to_embed_times_cos = op.Mul(to_embed, cos_4d)
    to_embed_x = op.Slice(to_embed, [0], [16], [3], [1])
    to_embed_y = op.Slice(to_embed, [16], [9223372036854775807], [3], [1])
    minus_to_embed_y = op.Neg(to_embed_y)
    to_embed_rotated_90 = op.Concat(minus_to_embed_y, to_embed_x, axis=-1)
    to_embed_rotated_90_times_sin = op.Mul(to_embed_rotated_90, sin_4d)
    embedded = op.Add(to_embed_times_cos, to_embed_rotated_90_times_sin)
    final = op.Concat(embedded, unembedded, axis=-1)
    return final


class _PartialRotaryTestCase:
    def get_onnx_model(self):
        if not hasattr(self, "_onnx_model"):
            model_proto = _partial_rotary_script.to_model_proto(
                input_types=(
                    INT64["Batchsize", "Sequence"],
                    FLOAT["Batchsize", 32, "Sequence", 80],
                ),
                output_types=(FLOAT["Batchsize", 32, "Sequence", 80],),
            )
            model = ir.serde.deserialize_model(model_proto)
            self._onnx_model = model
        return self._onnx_model

    def get_ort_inputs(self):
        if not hasattr(self, "_ort_inputs"):
            inputs = {
                "query": numpy.random.rand(1, 32, 8, 80).astype(numpy.float32),
                "position_ids": numpy.arange(8, dtype=numpy.int64).reshape(1, 8),
            }
            self._ort_inputs = inputs
        return self._ort_inputs


def partial_rotary_test_case():
    return _PartialRotaryTestCase()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Public API for values module.

This module re-exports the public API from the internal values module.
"""

from onnxscript._internal.values import (
    AttrRef,
    OnnxClosure,
    OnnxFunction,
    Op,
    Opset,
    SymbolValue,
    TracedOnnxFunction,
)

__all__ = [
    "AttrRef",
    "OnnxClosure",
    "OnnxFunction",
    "Op",
    "Opset",
    "SymbolValue",
    "TracedOnnxFunction",
]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""
A one-layer SmolLM model test case, with inputs: input_ids, attention_mask, and position_ids.
This is an onnxscript version of the model.
"""

import numpy as np
import onnx_ir as ir

from onnxscript import script
from onnxscript.onnx_opset import opset18
from onnxscript.onnx_types import FLOAT, INT64


def make_model(
    input_layernorm_weight_0,
    post_attention_layernorm_weight0,
    norm_weight,
    head_weight,
    self_attn_q_proj_weight0,
    self_attn_k_proj_weight0,
    self_attn_v_proj_weight0,
    self_attn_o_proj_weight0,
    mlp_gate_proj_weight0,
    mlp_up_proj_weight0,
    mlp_down_proj_weight0,
):
    @script()
    def main_graph(
        input0: INT64[1, 10], input1: FLOAT[1, 10], input2: INT64[1, 10]
    ) -> (FLOAT[1, 10, 49152], FLOAT[1, 32, 10, 64], FLOAT[1, 32, 10, 64]):
        model_layers_0_input_layernorm_weight = opset18.Constant(
            value=input_layernorm_weight_0
        )
        model_layers_0_post_attention_layernorm_weight = opset18.Constant(
            value=post_attention_layernorm_weight0
        )
        model_norm_weight = opset18.Constant(value=norm_weight)
        lm_head_weight = opset18.Constant(value=head_weight)
        model_layers_0_self_attn_q_proj_weight = opset18.Constant(
            value=self_attn_q_proj_weight0
        )
        model_layers_0_self_attn_k_proj_weight = opset18.Constant(
            value=self_attn_k_proj_weight0
        )
        model_layers_0_self_attn_v_proj_weight = opset18.Constant(
            value=self_attn_v_proj_weight0
        )
        model_layers_0_self_attn_o_proj_weight = opset18.Constant(
            value=self_attn_o_proj_weight0
        )
        model_layers_0_mlp_gate_proj_weight = opset18.Constant(value=mlp_gate_proj_weight0)
        model_layers_0_mlp_up_proj_weight = opset18.Constant(value=mlp_up_proj_weight0)
        model_layers_0_mlp_down_proj_weight = opset18.Constant(value=mlp_down_proj_weight0)

        embedding = opset18.Gather(lm_head_weight, input0, axis=0)
        minus_inf_10x10 = opset18.ConstantOfShape([10, 10], [-3.4028234663852886e38])
        mask_10x10 = opset18.Trilu(minus_inf_10x10, 1)
        slice_5 = opset18.Reshape(mask_10x10, [1, 1, 10, 10])
        unsqueeze_2 = opset18.Unsqueeze(input1, [1])
        unsqueeze_3 = opset18.Unsqueeze(unsqueeze_2, [2])
        add = slice_5 + unsqueeze_3
        eq = add == 0.0
        slice_10 = slice_5
        masked_fill = opset18.Where(eq, -3.4028235e38, slice_10)
        val_179 = opset18.Transpose(masked_fill, perm=[2, 1, 0, 3])
        slice_scatter = opset18.Transpose(val_179, perm=[2, 1, 0, 3])
        val_191 = opset18.Transpose(slice_scatter, perm=[1, 0, 2, 3])
        slice_scatter_1 = opset18.Transpose(val_191, perm=[1, 0, 2, 3])
        unsqueeze_6 = opset18.Unsqueeze(input2, [1])
        to_copy_1 = opset18.Cast(unsqueeze_6, to=1)
        view_1 = opset18.Constant(
            value=ir.tensor(
                np.array(
                    [
                        1.0,
                        0.7498942017555237,
                        0.5623413324356079,
                        0.4216965138912201,
                        0.3162277638912201,
                        0.23713736236095428,
                        0.17782793939113617,
                        0.1333521455526352,
                        0.10000000149011612,
                        0.07498941570520401,
                        0.05623412877321243,
                        0.04216964915394783,
                        0.03162277862429619,
                        0.0237137358635664,
                        0.017782794311642647,
                        0.01333521492779255,
                        0.009999999776482582,
                        0.007498942315578461,
                        0.005623413249850273,
                        0.0042169648222625256,
                        0.003162277862429619,
                        0.0023713738191872835,
                        0.0017782794311642647,
                        0.0013335214462131262,
                        0.0010000000474974513,
                        0.0007498941849917173,
                        0.000562341301701963,
                        0.00042169648804701865,
                        0.0003162277862429619,
                        0.0002371373848291114,
                        0.00017782794020604342,
                        0.0001333521504420787,
                    ],
                    dtype=np.float32,
                ).reshape([1, 32, 1])
            )
        )
        view_2 = opset18.Reshape(to_copy_1, [1, 1, 10], allowzero=0)
        bmm = view_1 @ view_2
        view_3 = opset18.Reshape(bmm, [1, 32, 10], allowzero=0)
        transpose = opset18.Transpose(view_3, perm=[0, 2, 1])
        cat = opset18.Concat(transpose, transpose, axis=-1)
        cos = opset18.Cos(cat)
        sin = opset18.Sin(cat)
        pow_1 = embedding**2.0
        mean = opset18.ReduceMean(pow_1, [-1], keepdims=1, noop_with_empty_axes=0)
        add_1 = mean + 1e-05
        val_244 = opset18.Sqrt(add_1)
        rsqrt = opset18.Reciprocal(val_244)
        mul_3 = embedding * rsqrt
        mul_4 = model_layers_0_input_layernorm_weight * mul_3
        t = opset18.Transpose(model_layers_0_self_attn_q_proj_weight, perm=[1, 0])
        view_5 = mul_4 @ t
        t_1 = opset18.Transpose(model_layers_0_self_attn_k_proj_weight, perm=[1, 0])
        view_7 = mul_4 @ t_1
        t_2 = opset18.Transpose(model_layers_0_self_attn_v_proj_weight, perm=[1, 0])
        view_9 = mul_4 @ t_2
        view_10 = opset18.Reshape(view_5, [1, 10, 32, 64], allowzero=0)
        transpose_1 = opset18.Transpose(view_10, perm=[0, 2, 1, 3])
        view_11 = opset18.Reshape(view_7, [1, 10, 32, 64], allowzero=0)
        transpose_2 = opset18.Transpose(view_11, perm=[0, 2, 1, 3])
        view_12 = opset18.Reshape(view_9, [1, 10, 32, 64], allowzero=0)
        transpose_3 = opset18.Transpose(view_12, perm=[0, 2, 1, 3])
        unsqueeze_7 = opset18.Unsqueeze(cos, [1])
        unsqueeze_8 = opset18.Unsqueeze(sin, [1])
        mul_5 = transpose_1 * unsqueeze_7
        val_267 = opset18.Constant(value_ints=[1])
        slice_19 = opset18.Slice(transpose_1, [0], [32], [3], val_267)
        val_277 = opset18.Constant(value_ints=[1])
        slice_20 = opset18.Slice(transpose_1, [32], [9223372036854775807], [3], val_277)
        neg = opset18.Neg(slice_20)
        cat_1 = opset18.Concat(neg, slice_19, axis=-1)
        mul_6 = cat_1 * unsqueeze_8
        add_2 = mul_5 + mul_6
        mul_7 = transpose_2 * unsqueeze_7
        val_287 = opset18.Constant(value_ints=[1])
        slice_21 = opset18.Slice(transpose_2, [0], [32], [3], val_287)
        val_297 = opset18.Constant(value_ints=[1])
        slice_22 = opset18.Slice(transpose_2, [32], [9223372036854775807], [3], val_297)
        neg_1 = opset18.Neg(slice_22)
        cat_2 = opset18.Concat(neg_1, slice_21, axis=-1)
        mul_8 = cat_2 * unsqueeze_8
        add_3 = mul_7 + mul_8
        val_346 = opset18.Reshape(add_3, [-1, 10, 64], allowzero=0)
        val_347 = opset18.Transpose(val_346, perm=[0, 2, 1])
        val_349 = opset18.Reshape(val_347, [1, 32, 64, 10], allowzero=0)
        val_351 = add_2 * [0.35355338]
        val_353 = val_349 * [0.35355338]
        val_354 = val_351 @ val_353
        val_355 = val_354 + slice_scatter_1
        val_356 = opset18.Softmax(val_355, axis=-1)
        getitem = val_356 @ transpose_3
        transpose_4 = opset18.Transpose(getitem, perm=[0, 2, 1, 3])
        view_13 = opset18.Reshape(transpose_4, [1, 10, -1], allowzero=0)
        t_3 = opset18.Transpose(model_layers_0_self_attn_o_proj_weight, perm=[1, 0])
        view_15 = view_13 @ t_3
        add_4 = embedding + view_15
        pow_2 = add_4**2.0
        mean_1 = opset18.ReduceMean(pow_2, [-1], keepdims=1, noop_with_empty_axes=0)
        add_5 = mean_1 + 1e-05
        val_379 = opset18.Sqrt(add_5)
        rsqrt_1 = opset18.Reciprocal(val_379)
        mul_9 = add_4 * rsqrt_1
        mul_10 = model_layers_0_post_attention_layernorm_weight * mul_9
        t_4 = opset18.Transpose(model_layers_0_mlp_gate_proj_weight, perm=[1, 0])
        view_17 = mul_10 @ t_4
        val_383 = opset18.Sigmoid(view_17)
        silu = view_17 * val_383
        t_5 = opset18.Transpose(model_layers_0_mlp_up_proj_weight, perm=[1, 0])
        view_19 = mul_10 @ t_5
        mul_11 = silu * view_19
        t_6 = opset18.Transpose(model_layers_0_mlp_down_proj_weight, perm=[1, 0])
        view_21 = mul_11 @ t_6
        add_6 = add_4 + view_21
        pow_3 = add_6**2.0
        mean_2 = opset18.ReduceMean(pow_3, [-1], keepdims=1, noop_with_empty_axes=0)
        add_7 = mean_2 + 1e-05
        val_391 = opset18.Sqrt(add_7)
        rsqrt_2 = opset18.Reciprocal(val_391)
        mul_12 = add_6 * rsqrt_2
        mul_13 = model_norm_weight * mul_12
        t_7 = opset18.Transpose(lm_head_weight, perm=[1, 0])
        view_23 = mul_13 @ t_7
        to_copy_12 = opset18.Identity(view_23)
        return to_copy_12, add_3, transpose_3

    model = main_graph.to_model_proto()
    return model


def make_model_with_random_weights():
    input_layernorm_weight_0 = np.random.rand(2048).astype(np.float32)
    post_attention_layernorm_weight0 = np.random.rand(2048).astype(np.float32)
    norm_weight = np.random.rand(2048).astype(np.float32)
    head_weight = np.random.rand(49152, 2048).astype(np.float32)
    self_attn_q_proj_weight0 = np.random.rand(2048, 2048).astype(np.float32)
    self_attn_k_proj_weight0 = np.random.rand(2048, 2048).astype(np.float32)
    self_attn_v_proj_weight0 = np.random.rand(2048, 2048).astype(np.float32)
    self_attn_o_proj_weight0 = np.random.rand(2048, 2048).astype(np.float32)
    mlp_gate_proj_weight0 = np.random.rand(8192, 2048).astype(np.float32)
    mlp_up_proj_weight0 = np.random.rand(8192, 2048).astype(np.float32)
    mlp_down_proj_weight0 = np.random.rand(2048, 8192).astype(np.float32)
    model = make_model(
        ir.tensor(input_layernorm_weight_0),
        ir.tensor(post_attention_layernorm_weight0),
        ir.tensor(norm_weight),
        ir.tensor(head_weight),
        ir.tensor(self_attn_q_proj_weight0),
        ir.tensor(self_attn_k_proj_weight0),
        ir.tensor(self_attn_v_proj_weight0),
        ir.tensor(self_attn_o_proj_weight0),
        ir.tensor(mlp_gate_proj_weight0),
        ir.tensor(mlp_up_proj_weight0),
        ir.tensor(mlp_down_proj_weight0),
    )
    return model


class _SmollmTest1:
    def get_onnx_model(self):
        if not hasattr(self, "_onnx_model"):
            model_proto = make_model_with_random_weights()
            model = ir.serde.deserialize_model(model_proto)
            self._onnx_model = model
        return self._onnx_model

    def get_ort_inputs(self):
        if not hasattr(self, "_ort_inputs"):
            inputs = {
                "input0": np.random.randint(0, 49152, (1, 10)).astype(np.int64),
                "input1": np.ones((1, 10), dtype=np.float32),
                "input2": np.arange(10, dtype=np.int64).reshape(1, 10),
            }
            self._ort_inputs = inputs
        return self._ort_inputs


def smollm_test_1():
    return _SmollmTest1()
